import json
from typing import Any
from typing import Dict


def convert_raw_bucket_notification_to_present(
    hub, raw_resource: Dict[str, Any], bucket_name: str
) -> Dict[str, Any]:
    """
    Convert the s3 bucket notification configurations response to a common format

    Args:
        raw_resource: Dictionary of s3 bucket notification configurations
        bucket_name: Name of the bucket on which notification needs to be configured.

    Returns:
        A dictionary of s3 bucket notification configurations
    """
    translated_resource = {}
    raw_resource.pop("ResponseMetadata", None)

    if raw_resource:
        translated_resource["name"] = bucket_name
        translated_resource["resource_id"] = bucket_name + "-notifications"
        translated_resource["notifications"] = json.loads(json.dumps(raw_resource))
    return translated_resource


def convert_raw_bucket_logging_to_present(
    hub, ctx, raw_resource: Dict[str, Any], bucket_name: str
) -> Dict[str, Any]:
    """
    Convert the s3 bucket logging configurations response to a common format

    Args:
        raw_resource: Dictionary of s3 bucket logging configurations
        bucket_name: Name of the bucket on which logging needs to be configured.

    Returns:
        A dictionary of s3 bucket logging configurations
    """
    translated_resource = {}
    raw_resource.pop("ResponseMetadata", None)

    if raw_resource:
        translated_resource["name"] = bucket_name
        translated_resource["resource_id"] = bucket_name
        translated_resource["bucket_logging_status"] = json.loads(
            json.dumps(raw_resource)
        )
    return translated_resource


async def convert_raw_s3_to_present(
    hub, ctx, idem_resource_name: str = None
) -> Dict[str, Any]:
    """
    Convert the s3 bucket configurations response to a common format

    Args:
        idem_resource_name: Name of the bucket.

    Returns:
        Dict[str, Any]
    """
    resource_translated = {
        "name": idem_resource_name,
        "resource_id": idem_resource_name,
    }
    bucket_acl: dict = {}
    result = dict(comment=(), result=True, ret=None)
    ret = await hub.exec.boto3.client.s3.get_object_lock_configuration(
        ctx, Bucket=idem_resource_name
    )
    if not ret["result"]:
        # Added error message check to continue if there is ObjectLockConfigurationNotFoundError exception else
        # return the result
        if "ObjectLockConfigurationNotFoundError" in str(ret["comment"]):
            result["comment"] = result["comment"] + ret["comment"]
        else:
            result["comment"] = result["comment"] + ret["comment"]
            result["result"] = False
            return result
    if ret["result"]:
        if ret["ret"]["ObjectLockConfiguration"]["ObjectLockEnabled"] == "Enabled":
            resource_translated["object_lock_enabled_for_bucket"] = True
            # Added the object_lock_configuration for put_object_lock_configuration operation.
            resource_translated["object_lock_configuration"] = json.loads(
                json.dumps(ret["ret"]["ObjectLockConfiguration"])
            )
        else:
            resource_translated["object_lock_enabled_for_bucket"] = False
    ret = await hub.exec.boto3.client.s3.get_bucket_location(
        ctx, Bucket=idem_resource_name
    )
    if not ret["result"]:
        result["comment"] = result["comment"] + ret["comment"]
        result["result"] = False
        return result
    if ret["result"] and ret["ret"]["LocationConstraint"]:
        resource_translated["create_bucket_configuration"] = {
            "LocationConstraint": ret["ret"]["LocationConstraint"]
        }
    ret = await hub.exec.boto3.client.s3.get_bucket_ownership_controls(
        ctx, Bucket=idem_resource_name
    )
    if not ret["result"]:
        # Added error message check to continue if there is OwnershipControlsNotFoundError exception else
        # return the result
        if "OwnershipControlsNotFoundError" in str(ret["comment"]):
            result["comment"] = result["comment"] + ret["comment"]
        else:
            result["comment"] = result["comment"] + ret["comment"]
            result["result"] = False
            return result
    if (
        ret["result"]
        and ret["ret"]["OwnershipControls"]
        and ret["ret"]["OwnershipControls"]["Rules"][0]["ObjectOwnership"]
    ):
        resource_translated["object_ownership"] = ret["ret"]["OwnershipControls"][
            "Rules"
        ][0]["ObjectOwnership"]
    ret = await hub.exec.boto3.client.s3.get_bucket_acl(ctx, Bucket=idem_resource_name)
    if not ret["result"]:
        # Added error message check to continue if there is AccessControlListNotSupported exception else
        # return the result
        if "AccessControlListNotSupported" in str(ret["comment"]):
            result["comment"] = result["comment"] + ret["comment"]
        else:
            result["comment"] = result["comment"] + ret["comment"]
            result["result"] = False
            return result
    # checking first element in grants because for no permissions API returns empty array
    full_control = ""
    read = ""
    read_acp = ""
    write = ""
    write_acp = ""

    if ret["result"] and ret["ret"]["Grants"][0]["Permission"]:
        for grant in ret["ret"]["Grants"]:
            if grant["Permission"]:
                grantee = grant["Grantee"]
                if grant["Permission"] == "FULL_CONTROL":
                    full_control = hub.tool.aws.s3.s3_utils.get_grantee_payload(
                        full_control, grantee
                    )
                elif grant["Permission"] == "READ":
                    read = hub.tool.aws.s3.s3_utils.get_grantee_payload(read, grantee)
                elif grant["Permission"] == "READ_ACP":
                    read_acp = hub.tool.aws.s3.s3_utils.get_grantee_payload(
                        read_acp, grantee
                    )
                elif grant["Permission"] == "WRITE":
                    write = hub.tool.aws.s3.s3_utils.get_grantee_payload(write, grantee)
                elif grant["Permission"] == "WRITE_ACP":
                    write_acp = hub.tool.aws.s3.s3_utils.get_grantee_payload(
                        write_acp, grantee
                    )

    if full_control:
        resource_translated["full_control"] = full_control

    if read:
        resource_translated["read"] = read

    if read_acp:
        resource_translated["read_acp"] = read_acp

    if write:
        resource_translated["write"] = write

    if write_acp:
        resource_translated["write_acp"] = write_acp

    ret = await hub.exec.boto3.client.s3.get_bucket_tagging(
        ctx, Bucket=idem_resource_name
    )
    if not ret["result"]:
        # Added error message check to continue if there is NoSuchTagSet exception else return the result
        if "NoSuchTagSet" in str(ret["comment"]):
            result["comment"] = result["comment"] + ret["comment"]
        else:
            result["comment"] = result["comment"] + ret["comment"]
            result["result"] = False
            return result
    if ret["result"] and ret["ret"]["TagSet"]:
        resource_translated["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            ret["ret"]["TagSet"]
        )
    result["ret"] = resource_translated

    return result


def convert_raw_public_access_block_to_present(
    hub, bucket_name: str, public_access_block_configuration: Dict
) -> Dict[str, Any]:
    translated_resource = {
        "name": f"{bucket_name}-public-access-block",
        "bucket": bucket_name,
        "public_access_block_configuration": json.loads(
            json.dumps(public_access_block_configuration)
        ),
        "resource_id": bucket_name,
    }
    return translated_resource


def convert_raw_bucket_policy_to_present(
    hub, raw_resource: Dict[str, Any], bucket: str, name: str
) -> Dict[str, Any]:
    """
    Util function to convert the s3 bucket policy response to a common format

    Args:
        hub: required for functions in hub
        raw_resource: S3 bucket policy response
        bucket: Name of the bucket on which policy needs to be configured.

    Returns:
        A dictionary of s3 bucket notification configurations
    """
    translated_resource = {}
    raw_resource.pop("ResponseMetadata", None)

    if raw_resource:
        translated_resource["name"] = name
        translated_resource["resource_id"] = bucket + "-policy"
        translated_resource["bucket"] = bucket
        translated_resource["policy"] = raw_resource["Policy"]
    return translated_resource


def convert_raw_bucket_replication_to_present(
    hub, ctx, raw_resource: Dict[str, Any], bucket_name: str
) -> Dict[str, Any]:
    """
    Convert the s3 bucket replication configurations response to a common format

    Args:
        raw_resource: Dictionary of s3 bucket replication configurations
        bucket_name: Name of the bucket on which replication needs to be configured.

    Returns:
        A dictionary of s3 bucket replication configurations
    """
    translated_resource = {}
    raw_resource.pop("ResponseMetadata", None)

    if raw_resource:
        translated_resource["name"] = bucket_name
        translated_resource["resource_id"] = bucket_name
        translated_resource["role"] = raw_resource["ReplicationConfiguration"]["Role"]
        translated_resource["rules"] = raw_resource["ReplicationConfiguration"]["Rules"]
    return translated_resource


def convert_raw_acl_policy_to_present(
    hub, raw_resource: Dict[str, Any], bucket_name: str
) -> Dict[str, Any]:
    """
    Util function to convert the s3 bucket policy response to a common format

    Args:
        hub: required for functions in hub
        raw_resource (Dict): S3 bucket policy response
        bucket_name(str): Name of the bucket on which policy needs to be configured.

    Returns:
        A dictionary of s3 bucket notification configurations
    """
    translated_resource = {}
    raw_resource.pop("ResponseMetadata", None)

    if raw_resource:
        translated_resource["resource_id"] = bucket_name
        translated_resource["name"] = bucket_name
        translated_resource["access_control_policy"] = json.loads(
            json.dumps(raw_resource)
        )

    return translated_resource



import copy
from typing import Any
from typing import Dict


async def update_eks_tags(
    hub,
    ctx,
    resource_arn,
    old_tags: Dict[str, Any],
    new_tags: Dict[str, Any],
):
    """

    Args:
        hub:
        ctx:
        resource_arn: aws resource arn
        old_tags: Dict of old tags
        new_tags: Dict of new tags

    Returns:
        {"result": True|False, "comment": "A message", "ret": None}

    """
    result = dict(comment=(), result=True, ret=None)

    tags_to_add = {}
    tags_to_remove = []
    tags_result = copy.deepcopy(old_tags)
    for key, value in new_tags.items():
        if (key in old_tags and old_tags.get(key) != new_tags.get(key)) or (
            key not in old_tags
        ):
            tags_to_add[key] = value

    for key in old_tags:
        if key not in new_tags:
            tags_to_remove.append(key)
    if tags_to_remove:
        if not ctx.get("test", False):
            delete_ret = await hub.exec.boto3.client.eks.untag_resource(
                ctx, resourceArn=resource_arn, tagKeys=tags_to_remove
            )
            if not delete_ret["result"]:
                result["comment"] = delete_ret["comment"]
                result["result"] = False
                return result
        [tags_result.pop(key.get("Key"), None) for key in tags_to_remove]
    if tags_to_add:
        if not ctx.get("test", False):
            add_ret = await hub.exec.boto3.client.eks.tag_resource(
                ctx, resourceArn=resource_arn, tags=tags_to_add
            )
            if not add_ret["result"]:
                result["comment"] = add_ret["comment"]
                result["result"] = False
                return result

    result["ret"] = {"tags": {**tags_result, **tags_to_add}}
    result["comment"] = (f"Update tags: Add [{tags_to_add}] Remove [{tags_to_remove}]",)
    return result



import copy
from typing import Any
from typing import Dict


async def update_tags(
    hub,
    ctx,
    resource_arn: str,
    old_tags: Dict[str, Any],
    new_tags: Dict[str, Any],
):
    """

    Args:
        hub:
        ctx:
        resource_arn: AWS resource arn
        old_tags: old tags in the format of {tag-key: tag-value}
        new_tags: new tags in the format of {tag-key: tag-value}. If this value is None, the function will do no operation on tags.

    Returns:
        {"result": True|False, "comment": A message Tuple, "ret": dict of updated tags}

    """
    result = dict(comment=(), result=True, ret={})

    tags_to_add = {}
    tags_to_delete = {}
    if new_tags is not None:
        tags_to_delete, tags_to_add = hub.tool.aws.tag_utils.diff_tags_dict(
            old_tags=old_tags, new_tags=new_tags
        )
    if (not tags_to_delete) and (not tags_to_add):
        result["ret"] = copy.deepcopy(old_tags if old_tags else {})
        return result

    if not ctx.get("test", False) and tags_to_delete:
        delete_tag_resp = await hub.exec.boto3.client.ecr.untag_resource(
            ctx, resourceArn=resource_arn, tagKeys=list(tags_to_delete.keys())
        )
        if not delete_tag_resp["result"]:
            hub.log.debug(
                f"Could not delete tags {tags_to_delete} for resource: '{resource_arn}' due to the error: {delete_tag_resp['comment']}"
            )
            result["comment"] = delete_tag_resp["comment"]
            result["result"] = False
            return result

    if not ctx.get("test", False) and tags_to_add:
        create_tag_resp = await hub.exec.boto3.client.ecr.tag_resource(
            ctx,
            resourceArn=resource_arn,
            tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags=tags_to_add),
        )
        if not create_tag_resp["result"]:
            hub.log.debug(
                f"Could not create tags {tags_to_add} for resource: '{resource_arn}' due to the error: {create_tag_resp['comment']}"
            )
            result["comment"] = create_tag_resp["comment"]
            result["result"] = False
            return result

    result["ret"] = new_tags
    if ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_update_tags_comment(
            tags_to_remove=tags_to_delete, tags_to_add=tags_to_add
        )
    else:
        result["comment"] = hub.tool.aws.comment_utils.update_tags_comment(
            tags_to_remove=tags_to_delete, tags_to_add=tags_to_add
        )

    return result



from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List


def convert_raw_load_balancer_to_present(
    hub,
    raw_resource: Dict[str, Any],
    idem_resource_name: str = None,
    tags: Dict[str, str] = None,
    attributes: List[Dict[str, Any]] = None,
) -> Dict[str, Any]:
    r"""
    Util function to convert raw resource state from AWS ElasticLoadBalancing Load Balancer to present input format.

    Args:
        raw_resource(Dict[str, Any]):The AWS response from 'ElasticLoadBalancing describe_load_balancers' to convert.
        idem_resource_name(Text, optional): An Idem name of the resource.
        tags(Dict[str, str], optional): The AWS ElasticLoadBalancing Load Balancer tags.
        attributes(List[Dict[str, Any]], optional):
            List of attributes associated with given ElasticLoadBalancing Load Balancer.

    Returns:
        Dict[str, Any]
    """

    resource_id = raw_resource.get("LoadBalancerName")
    resource_parameters = OrderedDict(
        {
            "AvailabilityZones": "availability_zones",
            "Subnets": "subnets",
            "SecurityGroups": "security_groups",
            "Scheme": "scheme",
            "Instances": "instances",
        }
    )
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    listeners = []
    if raw_resource.get("ListenerDescriptions"):
        for listener in raw_resource.get("ListenerDescriptions"):
            listeners.append(listener.get("Listener"))
    resource_translated["listeners"] = listeners
    if attributes:
        resource_translated["load_balancer_attributes"] = attributes
    if tags:
        resource_translated["tags"] = tags
    return resource_translated



from typing import Dict
from typing import List

from dict_tools import differ


async def update_nodegroup(
    hub,
    ctx,
    name: str,
    before: Dict,
    version: str,
    release_version: str,
    launch_template: Dict,
    labels: Dict,
    taints: List,
    scaling_config: Dict,
    update_config: Dict,
    client_request_token: Dict,
    timeout: Dict,
):
    """
    Updates an Amazon EKS node group

    Args:
       hub:
       ctx:
       name: An Idem name of the resource
       before(Dict): AWS cluster
       release_version(Text): The AMI version of the Amazon EKS optimized AMI to use with your node group.
                              By default, the latest available AMI version for the node group's current Kubernetes version is used. For
                              more information, see Amazon EKS optimized Amazon Linux 2 AMI versions in the Amazon EKS User Guide . If you
                              specify launchTemplate , and your launch template uses a custom AMI, then don't specify releaseVersion ,
                              or the node group deployment will fail. For more information about using launch templates with Amazon EKS,
                              see Launch template support in the Amazon EKS User Guide.

       version(Text): The desired Kubernetes version for your cluster
       launch_template(Dict): An object representing a node group's launch template specification. If
                              specified, then do not specify instanceTypes , diskSize , or remoteAccess and make sure that the launch
                              template meets the requirements in launchTemplateSpecification.
       update_config(Dict): The node group update configuration
       scaling_config(Dict): The scaling configuration details for the Auto Scaling group that is created
                                       for your node group
       taints(List): The Kubernetes taints to be applied to the nodes in the node group.
       labels(Dict): The Kubernetes labels to be applied to the nodes in the node group when they are created.
       client_request_token(Text):Unique, case-sensitive identifier that you provide to ensure the idempotency of the request.
                                  This field is autopopulated if not provided.
       timeout(Dict, optional): Timeout configuration for creating or updating nodegroup.
            * create (Dict) -- Timeout configuration for creating nodegroup
                * delay(int, default=30) -- The amount of time in seconds to wait between attempts.
                * max_attempts(int, default=80) -- Customized timeout configuration containing delay and max attempts.
            * update (string) -- Timeout configuration for updating nodegroup
                * delay(int, default=30) -- The amount of time in seconds to wait between attempts.
                * max_attempts(int, default=80) -- Customized timeout configuration containing delay and max attempts.

    Returns:
       {"result": True|False, "comment": A message Tuple, "ret": Dict}
    """
    result = dict(comment=(), result=True, ret={})

    # change depends on version, releaseVersion,
    update_version_payload = {}
    waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
        default_delay=30,
        default_max_attempts=80,
        timeout_config=timeout.get("update") if timeout else None,
    )
    if version and before.get("version") != version:
        update_version_payload["version"] = version
    if release_version and before.get("release_version") != release_version:
        update_version_payload["releaseVersion"] = release_version
    if update_version_payload:
        if not ctx.get("test", False):
            ret = await hub.exec.boto3.client.eks.update_nodegroup_version(
                ctx,
                nodegroupName=before["resource_id"],
                clusterName=before["cluster_name"],
                launchTemplate=launch_template,
                clientRequestToken=client_request_token,
                **update_version_payload,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                result["result"] = False
                return result
            else:
                hub.log.debug(f"Waiting on updating aws.eks.nodegroup '{name}'")
                try:
                    await hub.tool.boto3.client.wait(
                        ctx,
                        "eks",
                        "nodegroup_active",
                        nodegroupName=before.get("resource_id"),
                        clusterName=before.get("cluster_name"),
                        WaiterConfig=waiter_config,
                    )
                except Exception as e:
                    result["comment"] = result["comment"] + (str(e),)
                    result["result"] = False
                    return result
        if "version" in update_version_payload:
            result["ret"]["version"] = version
            result["comment"] = result["comment"] + (
                f"Update cluster nodegroup version: {version} on nodegroup {before.get('resource_id')}",
            )
        if "releaseVersion" in update_version_payload:
            result["ret"]["release_version"] = version
            result["comment"] = result["comment"] + (
                f"Update cluster nodegroup release_version: {release_version} on nodegroup {before.get('resource_id')}",
            )

    # change depends on labels, scaling_group, taint, update_config
    node_config_change = hub.tool.aws.eks.eks_utils.get_updated_node_group_config(
        before,
        labels,
        taints,
        scaling_config,
        update_config,
    )
    if node_config_change:
        if not ctx.get("test", False):
            ret = await hub.exec.boto3.client.eks.update_nodegroup_config(
                ctx,
                nodegroupName=before["resource_id"],
                clusterName=before["cluster_name"],
                labels=node_config_change.get("labels"),
                taints=node_config_change.get("taints"),
                scalingConfig=node_config_change.get("scaling_config"),
                updateConfig=node_config_change.get("update_config"),
                clientRequestToken=client_request_token,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = result["comment"] + ret["comment"]
                result["result"] = False
                return result
            else:
                hub.log.debug(f"Waiting on updating aws.eks.nodegroup '{name}'")
                try:
                    await hub.tool.boto3.client.wait(
                        ctx,
                        "eks",
                        "nodegroup_active",
                        nodegroupName=before.get("resource_id"),
                        clusterName=before.get("cluster_name"),
                        WaiterConfig=waiter_config,
                    )
                except Exception as e:
                    result["comment"] = result["comment"] + (str(e),)
                    result["result"] = False
                    return result
        for node_config_change_param in node_config_change:
            result["ret"][node_config_change_param] = node_config_change[
                node_config_change_param
            ]
            result["comment"] = result["comment"] + (
                f"Update cluster nodegroup {node_config_change_param}: {str(node_config_change[node_config_change_param])} on nodegroup {before.get('resource_id')}",
            )
    return result



import copy
from typing import Any
from typing import Dict
from typing import List


async def get_tags_for_resource(hub, ctx, resource_arn: str):
    """
    Gets the tags for a docdb resource

    Args:
        hub:
        ctx:
        resource_arn(string): aws resource arn

    Returns:
        {"result": True|False, "comment": "A message tuple", "ret": Dict['str', Any]|None}
    """
    result = dict(comment=(), result=False, ret=None)
    if not resource_arn:
        return result
    tags = await hub.exec.boto3.client.docdb.list_tags_for_resource(
        ctx, ResourceName=resource_arn
    )
    result["ret"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
        tags.get("ret").get("TagList") if tags.get("result") else None
    )
    result["result"] = tags.get("result")
    if not result["result"]:
        result[
            "comment"
        ] = f"Getting tags for resource with resource_arn: `{resource_arn}` failed"
    return result


async def update_tags(
    hub,
    ctx,
    resource_arn,
    old_tags: List[Dict[str, Any]],
    new_tags: List[Dict[str, Any]],
):
    """
    Update tags of AWS DocDb resources

    Args:
        hub:
        ctx:
        resource_arn: aws resource arn
        old_tags: list of old tags in format of [{"Key": tag-key, "Value": tag-value}]
        new_tags: list of new tags in format of [{"Key": tag-key, "Value": tag-value}]

    Returns:
        {"result": True|False, "comment": "A message tuple", "ret": Dict['str', any]}

    """
    tags_to_add = []
    tags_to_remove = []
    old_tags_map = {tag.get("Key"): tag for tag in old_tags or []}
    tags_result = copy.deepcopy(old_tags_map)
    if new_tags is not None:
        for tag in new_tags:
            if tag.get("Key") in old_tags_map:
                if tag.get("Value") == old_tags_map.get(tag.get("Key")).get("Value"):
                    del old_tags_map[tag.get("Key")]
                else:
                    tags_to_add.append(tag)
            else:
                tags_to_add.append(tag)
        tags_to_remove = list(old_tags_map.keys())
    result = dict(comment=(), result=True, ret=None)
    if (not tags_to_remove) and (not tags_to_add):
        return result
    if tags_to_remove:
        if not ctx.get("test", False):
            delete_ret = await hub.exec.boto3.client.docdb.remove_tags_from_resource(
                ctx, ResourceName=resource_arn, TagKeys=tags_to_remove
            )
            if not delete_ret["result"]:
                result["comment"] = delete_ret["comment"]
                result["result"] = False
                return result
        [tags_result.pop(key) for key in tags_to_remove]
    if tags_to_add:
        if not ctx.get("test", False):
            add_ret = await hub.exec.boto3.client.docdb.add_tags_to_resource(
                ctx, ResourceName=resource_arn, Tags=tags_to_add
            )
            if not add_ret["result"]:
                result["comment"] = add_ret["comment"]
                result["result"] = False
                return result
    result["ret"] = {
        "tags": hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            list(tags_result.values()) + tags_to_add
        )
    }
    result["comment"] = result["comment"] + (
        f"Update tags: Add [{tags_to_add}] Remove [{tags_to_remove}]",
    )
    return result



"""State module for managing S3 bucket's replication configurations."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List


__contracts__ = ["resource"]

TREQ = {
    "present": {"require": ["aws.s3.bucket.present"]},
}


async def present(
    hub,
    ctx,
    name: str,
    role: str,
    rules: List[
        make_dataclass(
            "ReplicationRule",
            [
                ("Status", str),
                (
                    "Destination",
                    make_dataclass(
                        "Destination",
                        [
                            ("Bucket", str),
                            ("Account", str, field(default=None)),
                            ("StorageClass", str, field(default=None)),
                            (
                                "AccessControlTranslation",
                                make_dataclass(
                                    "AccessControlTranslation",
                                    [("Owner", str)],
                                ),
                                field(default=None),
                            ),
                            (
                                "EncryptionConfiguration",
                                make_dataclass(
                                    "EncryptionConfiguration",
                                    [
                                        (
                                            "ReplicaKmsKeyID",
                                            str,
                                            field(default=None),
                                        )
                                    ],
                                ),
                                field(default=None),
                            ),
                            (
                                "ReplicationTime",
                                make_dataclass(
                                    "ReplicationTime",
                                    [
                                        ("Status", str),
                                        (
                                            "Time",
                                            make_dataclass(
                                                "ReplicationTimeValue",
                                                [
                                                    (
                                                        "Minutes",
                                                        int,
                                                        field(default=None),
                                                    )
                                                ],
                                            ),
                                        ),
                                    ],
                                ),
                                field(default=None),
                            ),
                            (
                                "Metrics",
                                make_dataclass(
                                    "Metrics",
                                    [
                                        ("Status", str),
                                        (
                                            "EventThreshold",
                                            "ReplicationTimeValue",
                                            field(default=None),
                                        ),
                                    ],
                                ),
                                field(default=None),
                            ),
                        ],
                    ),
                ),
                ("ID", str, field(default=None)),
                ("Priority", int, field(default=None)),
                ("Prefix", str, field(default=None)),
                (
                    "Filter",
                    make_dataclass(
                        "ReplicationRuleFilter",
                        [
                            ("Prefix", str, field(default=None)),
                            (
                                "Tag",
                                make_dataclass("Tag", [("Key", str), ("Value", str)]),
                                field(default=None),
                            ),
                            (
                                "And",
                                make_dataclass(
                                    "ReplicationRuleAndOperator",
                                    [
                                        (
                                            "Prefix",
                                            str,
                                            field(default=None),
                                        ),
                                        (
                                            "Tags",
                                            List["Tag"],
                                            field(default=None),
                                        ),
                                    ],
                                ),
                                field(default=None),
                            ),
                        ],
                    ),
                    field(default=None),
                ),
                (
                    "SourceSelectionCriteria",
                    make_dataclass(
                        "SourceSelectionCriteria",
                        [
                            (
                                "SseKmsEncryptedObjects",
                                make_dataclass(
                                    "SseKmsEncryptedObjects",
                                    [("Status", str)],
                                ),
                                field(default=None),
                            ),
                            (
                                "ReplicaModifications",
                                make_dataclass(
                                    "ReplicaModifications",
                                    [("Status", str)],
                                ),
                                field(default=None),
                            ),
                        ],
                    ),
                    field(default=None),
                ),
                (
                    "ExistingObjectReplication",
                    make_dataclass("ExistingObjectReplication", [("Status", str)]),
                    field(default=None),
                ),
                (
                    "DeleteMarkerReplication",
                    make_dataclass(
                        "DeleteMarkerReplication",
                        [("Status", str, field(default=None))],
                    ),
                    field(default=None),
                ),
            ],
        )
    ],
    resource_id: str = None,
) -> Dict[str, Any]:
    """Create/Update the replication configuration for S3 bucket.

    Creates a replication configuration or replaces an existing one. Specify the replication configuration in the
    request body. In the replication configuration, you provide the name of the destination bucket or buckets where you
    want Amazon S3 to replicate objects, the IAM role that Amazon S3 can assume to replicate objects on your behalf,
    and other relevant information. A replication configuration must include at least one rule, and can contain a
    maximum of 1,000. Each rule identifies a subset of objects to replicate by filtering the objects in the source bucket.
    To choose additional subsets of objects to replicate, add a rule for each subset. To specify a subset of the objects in
    the source bucket to apply a replication rule to, add the Filter element as a child of the Rule element. You can
    filter objects based on an object key prefix, one or more object tags, or both.

    Args:
        name(str):
            An Idem name of the resource.
            This is also used as the name of the bucket on which replication needs to be configured.

        resource_id(str, optional):
            Name of the S3 bucket.

        role (str):
            The Amazon Resource Name (ARN) of the Identity and Access Management (IAM) role that Amazon S3
            assumes when replicating objects.

        rules (list):
            A container for one or more replication rules. A replication configuration must have at least
            one rule and can contain a maximum of 1,000 rules.

            * ID (str, optional):
                A unique identifier for the rule. The maximum value is 255 characters.

            * Priority (int, optional):
                The priority indicates which rule has precedence whenever two or more replication rules
                conflict. Amazon S3 will attempt to replicate objects according to all replication rules.
                However, if there are two or more rules with the same destination bucket, then objects will be
                replicated according to the rule with the highest priority. The higher the number, the higher
                the priority.

            * Prefix (str, optional):
                An object key name prefix that identifies the object or objects to which the rule applies. The
                maximum prefix length is 1,024 characters. To include all objects in a bucket, specify an empty
                string. Replacement must be made for object keys containing special characters (such as
                carriage returns) when using XML requests.

            * Filter (dict, optional):
                A filter that identifies the subset of objects to which the replication rule applies. A Filter
                must specify exactly one Prefix, Tag, or an And child element.

                * Prefix (str, optional):
                    An object key name prefix that identifies the subset of objects to which the rule applies.
                    Replacement must be made for object keys containing special characters (such as carriage
                    returns) when using XML requests.

                * Tag (dict, optional):
                    A container for specifying a tag key and value.  The rule applies only to objects that have the
                    tag in their tag set.

                    * Key (str):
                        Name of the object key.

                    * Value (str):
                        Value of the tag.

                * And (dict, optional):
                    A container for specifying rule filters. The filters determine the subset of objects to which
                    the rule applies. This element is required only if you specify more than one filter. For
                    example: If you specify both a Prefix and a Tag filter, wrap these filters in an And tag.
                    If you specify a filter based on multiple tags, wrap the Tag elements in an And tag.

                    * Prefix (str, optional):
                        An object key name prefix that identifies the subset of objects to which the rule applies.

                    * Tags (list, optional):
                        An array of tags containing key and value pairs.
            * Status (str):
                Specifies whether the rule is enabled.

            * SourceSelectionCriteria (dict, optional):
                A container that describes additional filters for identifying the source objects that you want
                to replicate. You can choose to enable or disable the replication of these objects. Currently,
                Amazon S3 supports only the filter that you can specify for objects created with server-side
                encryption using a customer managed key stored in Amazon Web Services Key Management Service
                (SSE-KMS).

                * SseKmsEncryptedObjects (dict, optional):
                    A container for filter information for the selection of Amazon S3 objects encrypted with Amazon
                    Web Services KMS. If you include SourceSelectionCriteria in the replication configuration, this
                    element is required.

                    * Status (str):
                        Specifies whether Amazon S3 replicates objects created with server-side encryption using an
                        Amazon Web Services KMS key stored in Amazon Web Services Key Management Service.

                * ReplicaModifications (dict, optional):
                    A filter that you can specify for selections for modifications on replicas. Amazon S3 doesn't
                    replicate replica modifications by default. In the latest version of replication configuration
                    (when Filter is specified), you can specify this element and set the status to Enabled to
                    replicate modifications on replicas. If you don't specify the Filter element, Amazon S3
                    assumes that the replication configuration is the earlier version, V1. In the earlier version,
                    this element is not allowed

                    * Status (str):
                        Specifies whether Amazon S3 replicates modifications on replicas.

            * ExistingObjectReplication (dict, optional):

                * Status (str):

            * Destination (Dict[str, Any]):
                A container for information about the replication destination and its configurations including
                enabling the S3 Replication Time Control (S3 RTC).

                * Bucket (str):  The Amazon Resource Name (ARN) of the bucket where you want Amazon S3 to store the results.

                * Account (str, optional):
                    Destination bucket owner account ID. In a cross-account scenario, if you direct Amazon S3 to
                    change replica ownership to the Amazon Web Services account that owns the destination bucket by
                    specifying the AccessControlTranslation property, this is the account ID of the destination
                    bucket owner.

                * StorageClass (str, optional):
                    The storage class to use when replicating objects, such as S3 Standard or reduced redundancy.
                    By default, Amazon S3 uses the storage class of the source object to create the object replica.

                * AccessControlTranslation (dict, optional):
                    Specify this only in a cross-account scenario (where source and destination bucket owners are
                    not the same), and you want to change replica ownership to the Amazon Web Services account that
                    owns the destination bucket. If this is not specified in the replication configuration, the
                    replicas are owned by same Amazon Web Services account that owns the source object.

                    * Owner (str):
                        Specifies the replica ownership.

                * EncryptionConfiguration (dict, optional):
                    A container that provides information about encryption. If SourceSelectionCriteria is specified,
                    you must specify this element.

                    * ReplicaKmsKeyID (str, optional):
                        Specifies the ID (Key ARN or Alias ARN) of the customer managed Amazon Web Services KMS key
                        stored in Amazon Web Services Key Management Service (KMS) for the destination bucket. Amazon S3
                        uses this key to encrypt replica objects. Amazon S3 only supports symmetric, customer managed
                        KMS keys.

                * ReplicationTime (dict, optional):
                    A container specifying S3 Replication Time Control (S3 RTC), including whether S3 RTC is
                    enabled and the time when all objects and operations on objects must be replicated. Must be
                    specified together with a Metrics block.

                    * Status (str):
                        Specifies whether the replication time is enabled.

                    * Time (dict):
                        A container specifying the time by which replication should be complete for all objects and
                        operations on objects.

                        * Minutes (int, optional):
                            Contains an integer specifying time in minutes.   Valid value: 15

                * Metrics (dict, optional):
                    A container specifying replication metrics-related settings enabling replication metrics and
                    events.

                    * Status (str):
                        Specifies whether the replication metrics are enabled.

                    * EventThreshold ('ReplicationTimeValue', optional):
                        A container specifying the time threshold for emitting the s3:Replication:OperationMissedThreshold event.

            * DeleteMarkerReplication (dict, optional):
                Specifies whether Amazon S3 replicates delete markers. If you specify a Filter in your
                replication configuration, you must also include a DeleteMarkerReplication element. If your
                Filter includes a Tag element, the DeleteMarkerReplication Status must be set to Disabled,
                because Amazon S3 does not support replicating delete markers for tag-based rules. For an
                example configuration, see Basic Rule Configuration.  For more information about delete marker
                replication, see Basic Rule Configuration.   If you are using an earlier version of the
                replication configuration, Amazon S3 handles replication of delete markers differently.

                * Status (str, optional):
                    Indicates whether to replicate delete markers.  Indicates whether to replicate delete markers.

    Request Syntax:
        .. code-block:: sls

            [bucket_name]-replication:
              aws.s3.bucket_replication.present:
              - name: "string"
              - resource_id: "string"
              - replication_configuration:
                Role: "string"
                Rules:
                - DeleteMarkerReplication:
                    Status: "string"
                  Destination:
                    Bucket: "string"
                    EncryptionConfiguration:
                      ReplicaKmsKeyID: "string"
                    StorageClass: "string"
                  Filter:
                    And:
                      Prefix: "string"
                      Tags:
                      - Key: "string"
                        Value: "string"
                  ID: "string"
                  Priority: "int"
                  SourceSelectionCriteria:
                    SseKmsEncryptedObjects:
                      Status: "string"
                  Status: "string"

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_present:
              aws.s3.bucket_replication.present:
                - name: replication-test-bucket
                - replication_configuration:
                    Role: arn:aws:iam::460671877902:role/admin
                    Rules:
                    - DeleteMarkerReplication:
                        Status: Disabled
                      Destination:
                        Bucket: arn:aws:s3:::replication-bucket-demo-1
                        EncryptionConfiguration:
                          ReplicaKmsKeyID: arn:aws:kms:us-east-1:460671877902:alias/aws/s3
                        StorageClass: STANDARD
                      Filter:
                        And:
                          Prefix: logx
                          Tags:
                          - Key: name
                            Value: replication
                      ID: rule-1
                      Priority: 0
                      SourceSelectionCriteria:
                        SseKmsEncryptedObjects:
                          Status: Enabled
                      Status: Enabled
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    resource_updated = False

    if resource_id:
        before = await hub.exec.aws.s3.bucket_replication.get(
            ctx, resource_id=resource_id, name=name
        )
        if not before["result"] or not before["ret"]:
            result["result"] = False
            result["comment"] = before["comment"]
            return result

        result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
            resource_type="aws.s3.bucket_replication", name=name
        )
        result["old_state"] = copy.deepcopy(before["ret"])
        result["new_state"] = copy.deepcopy(result["old_state"])

        update_ret = await hub.tool.aws.s3.bucket_replication.update(
            ctx,
            resource_id=resource_id,
            before=before["ret"],
            role=role,
            rules=rules,
        )
        result["comment"] += update_ret["comment"]
        if not update_ret["result"]:
            result["result"] = False
            return result

        resource_updated = resource_updated or bool(update_ret["ret"])
        if update_ret["ret"] and ctx.get("test", False):
            result["new_state"].update(update_ret["ret"])
            result["comment"] += hub.tool.aws.comment_utils.would_update_comment(
                resource_type="aws.s3.bucket_replication", name=name
            )
        if resource_updated and ctx.get("test", False):
            return result
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "resource_id": name,
                    "role": role,
                    "rules": rules,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.s3.bucket_replication", name=name
            )
            return result
        payload = {"Role": role, "Rules": rules}
        create_ret = await hub.exec.boto3.client.s3.put_bucket_replication(
            ctx, Bucket=name, ReplicationConfiguration=payload
        )
        if not create_ret["result"]:
            result["result"] = False
            result["comment"] = create_ret["comment"]
            return result

        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.s3.bucket_replication", name=name
        )
        resource_id = name

    if (not result["old_state"]) or resource_updated:
        after_ret = await hub.exec.aws.s3.bucket_replication.get(
            ctx, name=name, resource_id=resource_id
        )
        if not after_ret["result"] or not after_ret["ret"]:
            result["result"] = False
            result["comment"] += tuple(after_ret["comment"])
            return result
        result["new_state"] = after_ret["ret"]

    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str,
) -> Dict[str, Any]:
    """Deletes the replication configuration for an S3 bucket.

    Deletes the replication configuration from the bucket. To use this operation, you must have permissions to
    perform the s3:PutReplicationConfiguration action. The bucket owner has these permissions by default and can
    grant it to others. It can take a while for the deletion of a replication configuration to fully propagate.

    Args:
        name(str): Idem name of the bucket.
        resource_id(str): Name of the bucket on which replication needs to be deleted.

    Request Syntax:
        .. code-block:: sls

            [bucket_name]-replication:
              aws.s3.bucket_replication.absent:
                - name: "string"
                - resource_id: "string"

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            test-bucket-replication:
              aws.s3.bucket_replication.absent:
                - name: bucket
                - resource_id: bucket
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.s3.bucket_replication", name=name
        )
        return result

    before = await hub.exec.aws.s3.bucket_replication.get(
        ctx, name=name, resource_id=resource_id
    )
    if not before["result"]:
        result["comment"] = before["comment"]
        result["result"] = False
        return result

    if not before["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.s3.bucket_replication", name=name
        )
    elif ctx.get("test", False):
        result["old_state"] = before["ret"]
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.s3.bucket_replication", name=name
        )
        return result
    else:
        result["old_state"] = before["ret"]
        delete_ret = await hub.exec.boto3.client.s3.delete_bucket_replication(
            ctx, Bucket=resource_id
        )
        if not delete_ret["result"]:
            result["result"] = False
            result["comment"] = delete_ret["comment"]
            return result

        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.s3.bucket_replication", name=name
        )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Gets information about the S3 bucket replication configurations.

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:
        .. code-block:: bash

            $ idem describe aws.s3.bucket_replication
    """
    result = {}
    # To describe the replication configurations of all the buckets, we first need to list all the buckets,
    # then get the replication configurations of each bucket
    ret = await hub.exec.boto3.client.s3.list_buckets(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not describe S3 buckets {ret['comment']}")
        return {}

    for bucket in ret["ret"]["Buckets"]:
        bucket_name = bucket.get("Name")
        # get replication configuration for each bucket
        bucket_replication_response = (
            await hub.exec.boto3.client.s3.get_bucket_replication(
                ctx, Bucket=bucket_name
            )
        )
        if bucket_replication_response["result"]:
            bucket_replication_response["ret"].pop("ResponseMetadata", None)
            if bucket_replication_response["ret"]:
                translated_resource = hub.tool.aws.s3.conversion_utils.convert_raw_bucket_replication_to_present(
                    ctx=ctx,
                    raw_resource=bucket_replication_response["ret"],
                    bucket_name=bucket_name,
                )

                result[bucket_name + "-replication"] = {
                    "aws.s3.bucket_replication.present": [
                        {parameter_key: parameter_value}
                        for parameter_key, parameter_value in translated_resource.items()
                    ]
                }
    return result



"""State module for managing Amazon Elasticache Parameter Groups."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    cache_parameter_group_family: str,
    description: str,
    resource_id: str = None,
    parameter_name_values: List[
        make_dataclass(
            """A list of parameter names and values for the parameter update."""
            "ParameterNameValue",
            [("ParameterName", str), ("ParameterValue", str, field(default=None))],
        )
    ] = None,
    tags: Dict[str, Any]
    or List[
        make_dataclass("Tag", [("Key", str), ("Value", str, field(default=None))])
    ] = None,
) -> Dict[str, Any]:
    """Creates an AWS Elasticache Parameter Group.

    Args:
        name(str):
            An Idem name of the resource
        resource_id(str, Optional):
            The name of the AWS Elasticache Parameter Group in Amazon Web Services.
        cache_parameter_group_family(str):
            The name of the cache parameter group family that the cache parameter group can be used with.
            Valid values are ``memcached1.4``, ``memcached1.5``, ``memcached1.6``, ``redis2.6``, ``redis2.8``,
            ``redis3.2``, ``redis4.0``, ``redis5.0``, and ``redis6.x``.
        description(str):
            A user-specified description for the cache parameter group.
        parameter_name_values(list, Optional):
            A list of parameter names and values for the parameter update.
            You must supply at least one parameter name and value; subsequent arguments are optional.
            A maximum of 20 parameters may be modified per request.

            * ParameterName (*str, Optional*):
                The name of the parameter.

            * ParameterValue (*str, Optional*):
                The value of the parameter.

        tags(dict or list, Optional):
            Dict in the format of ``{tag-key: tag-value}`` or List of tags in the format of
            ``[{"Key": tag-key, "Value": tag-value}]`` to associate with the parameter group.

            * Key (*str*):
                The key of the tag.
            * Value (*str*):
                The value of the tag.

    Request Syntax:
      .. code-block:: sls

        [idem_test_aws_elasticache_cache_parameter_group]:
          aws.elasticache.cache_parameter_group.present:
            - name: 'string'
            - resource_id: 'string'
            - cache_parameter_group_family: 'memcached1.4|memcached1.5|memcached1.6|redis2.6|redis2.8|redis3.2|redis4.0|redis5.0|redis6.x'
            - description: 'string'
            - parameter_name_values:
                - ParameterName: 'string'
                  ParameterValue: 'string'
            - tags:
                - Key: 'string'
                  Value: 'string'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            idem_test_aws_elasticache_cache_parameter_group:
              aws.elasticache.cache_parameter_group.present:
                - name: 'idem_test_cache_parameter_group'
                - cache_parameter_group_family: 'redis6.x'
                - description: 'My Elasticache Parameter Group'
                - parameter_name_values:
                  - ParameterName: 'acllog-max-len'
                    ParameterValue: '128'
                - tags:
                    - Key: 'provider'
                      Value: 'idem'
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    if resource_id:
        before = (
            await hub.exec.boto3.client.elasticache.describe_cache_parameter_groups(
                ctx, CacheParameterGroupName=resource_id
            )
        )
    if isinstance(tags, List):
        tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)

    is_updated = False
    are_tags_updated = False
    old_cache_parameter_group = {}
    resource_arn = str
    parameters = []

    if before and before["result"]:
        try:
            convert_ret = await hub.tool.aws.elasticache.conversion_utils.convert_raw_elasticache_parameter_group_to_present(
                ctx=ctx,
                raw_resource=before["ret"]["CacheParameterGroups"][0],
                idem_resource_name=resource_id,
            )
            result["result"] = convert_ret["result"]
            if not result["result"]:
                result["comment"] = result["comment"] + convert_ret["comment"]
            result["old_state"] = convert_ret["ret"]

            old_cache_parameter_group = copy.deepcopy(result["old_state"])
            plan_state = old_cache_parameter_group
            parameters = result["old_state"].get("parameter_name_values")
            resource_arn = result["old_state"].get("arn")
            result["comment"] = (
                f"aws.elasticache.cache_parameter_group '{name}' already exists",
            )

            # Update tags
            if tags is not None and tags != result["old_state"].get("tags"):
                update_tags_ret = (
                    await hub.tool.aws.elasticache.elasticache_utils.update_tags(
                        ctx=ctx,
                        resource_arn=resource_arn,
                        old_tags=result["old_state"].get("tags"),
                        new_tags=tags,
                    )
                )
                if not update_tags_ret["result"]:
                    result["comment"] = result["comment"] + update_tags_ret["comment"]
                    result["result"] = False
                    return result
                are_tags_updated = bool(update_tags_ret["result"])
                if ctx.get("test", False) and are_tags_updated:
                    plan_state["tags"] = update_tags_ret["ret"]
                    result["comment"] = result["comment"] + (
                        f"Would update tags for aws.elasticache.cache_parameter_group {name}",
                    )
        except Exception as e:
            result["comment"] = result["comment"] + (str(e),)
            result["result"] = False
            return result
    else:
        try:
            if ctx.get("test", False):
                result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                    enforced_state={},
                    desired_state={
                        "name": name,
                        "cache_parameter_group_family": cache_parameter_group_family,
                        "description": description,
                        "parameter_name_values": parameter_name_values,
                        "tags": tags,
                    },
                )
                result["comment"] = result["comment"] + (
                    f"Would create aws.elasticache.cache_parameter_group '{name}'",
                )
                return result

            ret = await hub.exec.boto3.client.elasticache.create_cache_parameter_group(
                ctx,
                CacheParameterGroupName=name,
                CacheParameterGroupFamily=cache_parameter_group_family,
                Description=description,
                Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
                if tags
                else None,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            resource_id = name
            after_parameters = (
                await hub.exec.boto3.client.elasticache.describe_cache_parameters(
                    ctx, CacheParameterGroupName=resource_id
                )
            )
            if not after_parameters["result"]:
                result["comment"] = result["comment"] + after_parameters["comment"]
                result["result"] = False
                return result
            parameters = after_parameters["ret"].get("Parameters")
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False
            return result

    # To add,remove,update the ParameterNameValues.
    # ParameterNameValues are defined automatically when a resource is created,
    # so this needs to be done regardless of whether the resource exists before.
    if parameter_name_values:
        update_parameter = (
            await hub.tool.aws.elasticache.parameters.update_elasticache_parameters(
                ctx=ctx,
                resource_id=resource_id,
                old_parameters=parameters,
                parameter_name_values=parameter_name_values,
            )
        )
        if not update_parameter["result"]:
            result["comment"] = result["comment"] + update_parameter["comment"]
            result["result"] = False
            return result
        is_updated = update_parameter["is_updated"]
        if is_updated:
            after_parameters = (
                await hub.exec.boto3.client.elasticache.describe_cache_parameters(
                    ctx, CacheParameterGroupName=resource_id
                )
            )
            if not after_parameters["result"]:
                result["comment"] = result["comment"] + after_parameters["comment"]
                result["result"] = False
                return result
            parameters = after_parameters.get("ret").get("Parameters")
        if ctx.get("test", False) and is_updated:
            plan_state["parameter_name_values"] = parameters

    try:
        if ctx.get("test", False):
            new_cache_parameter_group = plan_state
        elif (not (before and before["result"])) or is_updated or are_tags_updated:
            after = (
                await hub.exec.boto3.client.elasticache.describe_cache_parameter_groups(
                    ctx, CacheParameterGroupName=resource_id
                )
            )
            if after and after.get("ret"):
                convert_ret = await hub.tool.aws.elasticache.conversion_utils.convert_raw_elasticache_parameter_group_to_present(
                    ctx=ctx,
                    raw_resource=after["ret"]["CacheParameterGroups"][0],
                    idem_resource_name=resource_id,
                )
                result["result"] = convert_ret["result"]
                if not result["result"]:
                    result["comment"] = result["comment"] + convert_ret["comment"]
                new_cache_parameter_group = convert_ret.get("ret")
                if not (before and before["result"]):
                    result["comment"] = result["comment"] + (
                        f"Created aws.elasticache.cache_parameter_group '{name}'",
                    )
                elif is_updated or are_tags_updated:
                    result["comment"] = result["comment"] + (
                        f"Updated aws.elasticache.cache_parameter_group '{name}'",
                    )
            else:
                result["result"] = result["result"] and after["result"]
                if not result["result"]:
                    result["comment"] = result["comment"] + after["comment"]
                    return result
        else:
            new_cache_parameter_group = copy.deepcopy(old_cache_parameter_group)

        result["old_state"] = old_cache_parameter_group
        result["new_state"] = new_cache_parameter_group

    except Exception as e:
        result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
        result["result"] = False
    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    """Deletes the specified AWS Elasticache Parameter Group.

    .. warning::
        You cannot delete a cache parameter group if it is associated with any cache clusters.
        You cannot delete the default cache parameter groups in your account.

    Args:
        name(str):
            The name of the AWS Elasticache Parameter Group in Amazon Web Services.
        resource_id(str, Optional):
            The name of the AWS Elasticache Parameter Group in Amazon Web Services.

    Request Syntax:
      .. code-block:: sls

        [idem_test_aws_elasticache_cache_parameter_group]:
          aws.elasticache.cache_parameter_group.absent:
            - name: 'string'
            - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            idem_test_aws_elasticache_cache_parameter_group:
              aws.elasticache.cache_parameter_group.absent:
                - name: 'idem_test_cache_parameter_group'
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    before = await hub.exec.boto3.client.elasticache.describe_cache_parameter_groups(
        ctx,
        CacheParameterGroupName=name,
    )
    if not (before and before["result"]):
        result["comment"] = (
            f"aws.elasticache.cache_parameter_group '{name}' already absent",
        )
        return result
    else:
        convert_ret = await hub.tool.aws.elasticache.conversion_utils.convert_raw_elasticache_parameter_group_to_present(
            ctx=ctx,
            raw_resource=before["ret"]["CacheParameterGroups"][0],
            idem_resource_name=name,
        )
        result["result"] = convert_ret["result"]
        if not result["result"]:
            result["comment"] = result["comment"] + convert_ret["comment"]
        result["old_state"] = convert_ret["ret"]
        if ctx.get("test", False):
            result["comment"] = (
                f"Would delete aws.elasticache.cache_parameter_group '{name}'",
            )
            return result
        else:
            try:
                ret = await hub.exec.boto3.client.elasticache.delete_cache_parameter_group(
                    ctx,
                    CacheParameterGroupName=name,
                )
                result["result"] = ret["result"]
                if not result["result"]:
                    result["comment"] = ret["comment"]
                    return result
                result["comment"] = (
                    f"Deleted aws.elasticache.cache_parameter_group '{name}'",
                )
            except hub.tool.boto3.exception.ClientError as e:
                result["comment"] = result["comment"] + (
                    f"{e.__class__.__name__}: {e}",
                )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describes Elasticache Parameter Groups in a way that can be recreated/managed with the corresponding "present" function.

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:
        .. code-block:: bash

            $ idem describe aws.elasticache.cache_parameter_group
    """
    result = {}
    ret = await hub.exec.boto3.client.elasticache.describe_cache_parameter_groups(ctx)
    if not ret["result"]:
        hub.log.debug(
            f"Could not describe Elasticache parameter groups {ret['comment']}"
        )
        return {}
    for cache_parameter_group in ret["ret"]["CacheParameterGroups"]:
        resource_id = cache_parameter_group.get("CacheParameterGroupName")
        convert_ret = await hub.tool.aws.elasticache.conversion_utils.convert_raw_elasticache_parameter_group_to_present(
            ctx=ctx,
            raw_resource=cache_parameter_group,
            idem_resource_name=resource_id,
        )
        if not convert_ret["result"]:
            hub.log.warning(
                f"Could not describe elasticache parameter group '{resource_id}' with error {convert_ret['comment']}"
            )
            continue
        resource_translated = convert_ret.get("ret")
        result[resource_id] = {
            "aws.elasticache.cache_parameter_group.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }
    return result



import copy
from typing import Any
from typing import Dict

from dict_tools import data
from dict_tools import differ


async def update_options(
    hub,
    ctx,
    transit_gateway_id: str,
    old_description: str,
    new_description: str,
    old_options: Dict[str, Any],
    new_options: Dict[str, Any],
):
    """
    Update options and description of a transit_gateway. This function compares the existing(old) options or description
    and the new options or description. TransitGatewayCidrBlocks that are in the new options but not in the old
    TransitGatewayCidrBlocks will be associated to transit gateway. TransitGatewayCidrBlocks that are in the
    options but not in the new options will be disassociated from transit gateway.

    Args:
        hub:
        ctx:
        transit_gateway_id(Text): The AWS resource id of the existing transit gateway
        old_description(string):  description of existing transit gateway
        new_description(string): new description to update
        old_options(dict): options of existing transit gateway
        new_options(dict): options to be updated

    Returns:
        {"result": True|False, "comment": "A message", "ret": None}

    """
    result = dict(comment=(), result=True, ret=None)
    resolved_new_options = copy.deepcopy(old_options)

    for key in new_options.keys():
        resolved_new_options[key] = new_options[key]

    # compare option
    diff_in_options = data.recursive_diff(
        old_options,
        resolved_new_options,
        ignore_order=True,
    )

    if old_description == new_description and (not diff_in_options):
        result["comment"] = (f"no changes to update {transit_gateway_id}",)
        return result
    modify_options = None
    if new_options is not None:
        modify_supported_options = [
            "VpnEcmpSupport",
            "DnsSupport",
            "AutoAcceptSharedAttachments",
            "DefaultRouteTableAssociation",
            "AssociationDefaultRouteTableId",
            "DefaultRouteTablePropagation",
            "PropagationDefaultRouteTableId",
        ]
        modify_options = {}
        for modify_option in modify_supported_options:
            if new_options.get(modify_option) is not None:
                modify_options[modify_option] = new_options.get(modify_option)

        old_cidr_blocks = old_options.get("TransitGatewayCidrBlocks", [])
        new_cidr_blocks = new_options.get("TransitGatewayCidrBlocks", [])
        modify_options["AddTransitGatewayCidrBlocks"] = list(
            set(new_cidr_blocks).difference(old_cidr_blocks)
        )
        modify_options["RemoveTransitGatewayCidrBlocks"] = list(
            set(old_cidr_blocks).difference(new_cidr_blocks)
        )
    if not ctx.get("test", False):
        ret = await hub.exec.boto3.client.ec2.modify_transit_gateway(
            ctx,
            TransitGatewayId=transit_gateway_id,
            Description=new_description,
            Options=modify_options,
        )
        if not ret.get("result"):
            result["comment"] = ret["comment"]
            result["result"] = False
            return result
    result["comment"] = (f"Updated transit gateway {transit_gateway_id}",)
    result["ret"] = {"description": new_description, "options": new_options}

    return result


async def get_transit_gateway_by_id(hub, ctx, transit_gateway_id):
    ret = {}
    result = dict(comment=(), result=True, ret=None)
    before = await hub.exec.boto3.client.ec2.describe_transit_gateways(
        ctx, TransitGatewayIds=[transit_gateway_id]
    )
    result["result"] = before["result"]
    result["comment"] = before["comment"]
    if before["result"]:
        for transit_gateway in before["ret"]["TransitGateways"]:
            ret = transit_gateway
    result["ret"] = ret
    return result



from typing import Dict
from typing import List


async def apply(
    hub,
    ctx,
    resource,
    *,
    old_value: List[Dict[str, str]],
    new_value: List[Dict[str, str]],
    comments: List[str],
) -> bool:
    """
    Modify an ec2 instance based on a single parameter in it's "present" state

    Args:
        hub:
        ctx: The ctx from a state module call
        resource: An ec2 instance resource object
        old_value: The previous value from the attributes of an existing instance
        new_value: The desired value from the ec2 instance present state parameters
        comments: A running list of comments abound the update process
    """
    if new_value is None:
        ret = await hub.exec.boto3.client.ec2.reset_instance_attribute(
            ctx, InstanceId=resource.id, Attribute="blockDeviceMapping"
        )
        if ret.comment:
            comments.append(ret.comment)
        return ret.result
    else:
        ret = await hub.exec.boto3.client.ec2.modfiy_instance_attribute(
            ctx,
            InstanceId=resource.id,
            Attribute="blockDeviceMapping",
            BlockDeviceMappings=new_value,
        )
        if ret.comment:
            comments.append(ret.comment)
        return ret.result



import copy
from typing import Any
from typing import Dict
from typing import List


async def get(
    hub,
    ctx,
    resource_id: str,
    name: str = None,
):
    # TODO: will revisit this part later to improve building the get() call when multiple params are used to fetch a resource from AWS
    """
    Get an API Gateway Method from AWS.
    Args:
        hub:
        ctx:
        resource_id(string): Idem Resource ID that is generated once the resource is created.
        name(string, optional): An Idem name of the API Gateway Method.

    """

    result = dict(comment=[], ret=None, result=True)
    get_method = await hub.exec.boto3.client.apigateway.get_method(
        ctx,
        restApiId=resource_id.split("-")[0],
        resourceId=resource_id.split("-")[1],
        httpMethod=resource_id.split("-")[2],
    )

    if not get_method["result"]:
        if "NotFoundException" in str(get_method["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.apigateway.method", name=name
                )
            )
            result["comment"] += list(get_method["comment"])
            return result
        result["comment"] += list(get_method["comment"])
        result["result"] = False
        return result
    if not get_method["ret"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.apigateway.method", name=name
            )
        )
        return result

    if get_method["ret"]:
        result["ret"] = hub.tool.aws.apigateway.method.convert_raw_method_to_present(
            raw_resource=get_method["ret"],
            idem_resource_name=name,
            resource_id=resource_id,
        )

    return result


async def update(
    hub,
    ctx,
    old_state: Dict[str, Any],
    update_parameters: Dict[str, Any],
):
    r"""

    Updates an API Gateway Method

    Args:
        hub:
        ctx:
        old_state(Dict): Previous state of resource
        update_parameters(Dict): Parameters from SLS File

    Returns:
        Dict[str, Any]

    """

    result = dict(comment=(), result=True, ret=None)
    patch_ops = []

    new_state = copy.deepcopy(old_state)
    updated = False

    for key, value in update_parameters.items():
        if key == "api_key_required":
            key_final = "apiKeyRequired"
        elif key == "authorization_type":
            key_final = "authorizationType"
        elif key == "http_method":
            key_final = "httpMethod"

        if value != old_state.get(key):
            patch_ops.append(
                {
                    "op": "replace",
                    "path": f"/{key_final}",
                    "value": str(value)
                    if not isinstance(value, List)
                    else ",".join(value),
                }
            )
            updated = True
        elif old_state.get(key) is None:
            patch_ops.append(
                {
                    "op": "add",
                    "path": f"/{key_final}",
                    "value": str(value)
                    if not isinstance(value, List)
                    else ",".join(value),
                }
            )
            new_state[key] = value
            updated = True

    if ctx.get("test", False):
        if updated:
            result["comment"] = hub.tool.aws.comment_utils.would_update_comment(
                resource_type="aws.apigateway.method", name=old_state["name"]
            )
            result["ret"] = new_state
        else:
            result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
                resource_type="aws.apigateway.method", name=old_state["name"]
            )
        return result

    if updated:
        update_ret = await hub.exec.boto3.client.apigateway.update_method(
            ctx,
            restApiId=old_state["rest_api_id"],
            resourceId=old_state["parent_resource_id"],
            httpMethod=old_state["http_method"],
            patchOperations=patch_ops,
        )
        if not update_ret["result"]:
            result["result"] = False
            result["comment"] = update_ret["comment"]
            return result
        else:
            result["comment"] = hub.tool.aws.comment_utils.update_comment(
                resource_type="aws.apigateway.method", name=old_state["name"]
            )
            result["ret"] = new_state
    else:
        result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
            resource_type="aws.apigateway.method", name=old_state["name"]
        )

    return result



from typing import Dict


def create_waiter_config(
    hub, default_delay: int, default_max_attempts: int, timeout_config: Dict or None
) -> Dict[str, int]:
    """
    Create a waiter configuration that can be used as input in boto3 waiter calls. The configuration is based on
    the default values and the customized value in timeout_config.

    Args:
        hub: The redistributed pop central hub.
        default_delay: The amount of time in seconds to wait between attempts.
        default_max_attempts: The maximum number of attempts to be made.
        timeout_config: Customized timeout configuration containing delay and max attempts.

    Returns:
        {"Delay": delay-time, "MaxAttempts": max-attempts}
    """
    result = {"Delay": default_delay, "MaxAttempts": default_max_attempts}
    if timeout_config:
        if "delay" in timeout_config:
            result["Delay"] = timeout_config["delay"]
        if "max_attempts" in timeout_config:
            result["MaxAttempts"] = timeout_config["max_attempts"]
    return result



"""State module for managing Scaling policies."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    auto_scaling_group_name: str,
    resource_id: str = None,
    policy_type: str = None,
    adjustment_type: str = None,
    min_adjustment_step: int = None,
    min_adjustment_magnitude: int = None,
    scaling_adjustment: int = None,
    cooldown: int = None,
    metric_aggregation_type: str = None,
    step_adjustments: List[
        make_dataclass(
            "StepAdjustment",
            [
                ("ScalingAdjustment", int),
                ("MetricIntervalLowerBound", float, field(default=None)),
                ("MetricIntervalUpperBound", float, field(default=None)),
            ],
        )
    ] = None,
    estimated_instance_warmup: int = None,
    target_tracking_configuration: make_dataclass(
        "TargetTrackingConfiguration",
        [
            ("TargetValue", float),
            (
                "PredefinedMetricSpecification",
                make_dataclass(
                    "PredefinedMetricSpecification",
                    [
                        ("PredefinedMetricType", str),
                        ("ResourceLabel", str, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
            (
                "CustomizedMetricSpecification",
                make_dataclass(
                    "CustomizedMetricSpecification",
                    [
                        ("MetricName", str),
                        ("Namespace", str),
                        ("Statistic", str),
                        (
                            "Dimensions",
                            List[
                                make_dataclass(
                                    "MetricDimension", [("Name", str), ("Value", str)]
                                )
                            ],
                            field(default=None),
                        ),
                        ("Unit", str, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
            ("DisableScaleIn", bool, field(default=None)),
        ],
    ) = None,
    enabled: bool = None,
    predictive_scaling_configuration: make_dataclass(
        "PredictiveScalingConfiguration",
        [
            (
                "MetricSpecifications",
                List[
                    make_dataclass(
                        "PredictiveScalingMetricSpecification",
                        [
                            ("TargetValue", float),
                            (
                                "PredefinedMetricPairSpecification",
                                make_dataclass(
                                    "PredictiveScalingPredefinedMetricPair",
                                    [
                                        ("PredefinedMetricType", str),
                                        ("ResourceLabel", str, field(default=None)),
                                    ],
                                ),
                                field(default=None),
                            ),
                            (
                                "PredefinedScalingMetricSpecification",
                                make_dataclass(
                                    "PredictiveScalingPredefinedScalingMetric",
                                    [
                                        ("PredefinedMetricType", str),
                                        ("ResourceLabel", str, field(default=None)),
                                    ],
                                ),
                                field(default=None),
                            ),
                            (
                                "PredefinedLoadMetricSpecification",
                                make_dataclass(
                                    "PredictiveScalingPredefinedLoadMetric",
                                    [
                                        ("PredefinedMetricType", str),
                                        ("ResourceLabel", str, field(default=None)),
                                    ],
                                ),
                                field(default=None),
                            ),
                            (
                                "CustomizedScalingMetricSpecification",
                                make_dataclass(
                                    "PredictiveScalingCustomizedScalingMetric",
                                    [
                                        (
                                            "MetricDataQueries",
                                            List[
                                                make_dataclass(
                                                    "MetricDataQuery",
                                                    [
                                                        ("Id", str),
                                                        (
                                                            "Expression",
                                                            str,
                                                            field(default=None),
                                                        ),
                                                        (
                                                            "MetricStat",
                                                            make_dataclass(
                                                                "MetricStat",
                                                                [
                                                                    (
                                                                        "Metric",
                                                                        make_dataclass(
                                                                            "Metric",
                                                                            [
                                                                                (
                                                                                    "Namespace",
                                                                                    str,
                                                                                ),
                                                                                (
                                                                                    "MetricName",
                                                                                    str,
                                                                                ),
                                                                                (
                                                                                    "Dimensions",
                                                                                    List[
                                                                                        make_dataclass(
                                                                                            "MetricDimension",
                                                                                            [
                                                                                                (
                                                                                                    "Name",
                                                                                                    str,
                                                                                                ),
                                                                                                (
                                                                                                    "Value",
                                                                                                    str,
                                                                                                ),
                                                                                            ],
                                                                                        )
                                                                                    ],
                                                                                    field(
                                                                                        default=None
                                                                                    ),
                                                                                ),
                                                                            ],
                                                                        ),
                                                                    ),
                                                                    ("Stat", str),
                                                                    (
                                                                        "Unit",
                                                                        str,
                                                                        field(
                                                                            default=None
                                                                        ),
                                                                    ),
                                                                ],
                                                            ),
                                                            field(default=None),
                                                        ),
                                                        (
                                                            "Label",
                                                            str,
                                                            field(default=None),
                                                        ),
                                                        (
                                                            "ReturnData",
                                                            bool,
                                                            field(default=None),
                                                        ),
                                                    ],
                                                )
                                            ],
                                        )
                                    ],
                                ),
                                field(default=None),
                            ),
                            (
                                "CustomizedLoadMetricSpecification",
                                make_dataclass(
                                    "PredictiveScalingCustomizedLoadMetric",
                                    [("MetricDataQueries", List["MetricDataQuery"])],
                                ),
                                field(default=None),
                            ),
                            (
                                "CustomizedCapacityMetricSpecification",
                                make_dataclass(
                                    "PredictiveScalingCustomizedCapacityMetric",
                                    [("MetricDataQueries", List["MetricDataQuery"])],
                                ),
                                field(default=None),
                            ),
                        ],
                    )
                ],
            ),
            ("Mode", str, field(default=None)),
            ("SchedulingBufferTime", int, field(default=None)),
            ("MaxCapacityBreachBehavior", str, field(default=None)),
            ("MaxCapacityBuffer", int, field(default=None)),
        ],
    ) = None,
) -> Dict[str, Any]:
    r"""Creates or updates a scaling policy for an Auto Scaling group.

    Scaling policies are used to scale an Auto Scaling group based on configurable metrics. If no policies are defined,
    the dynamic scaling and predictive scaling features are not used.

    For more information about using dynamic scaling, see Target tracking scaling policies and Step and simple scaling
    policies in the Amazon EC2 Auto Scaling User Guide.

    For more information about using predictive scaling, see Predictive scaling for Amazon EC2 Auto Scaling in the
    Amazon EC2 Auto Scaling User Guide.

    You can view the scaling policies for an Auto Scaling group using the DescribePolicies API
    call. If you are no longer using a scaling policy, you can delete it by calling the DeletePolicy API.

    Args:
        name(str):
            The name of the scaling policy.

        auto_scaling_group_name(str):
            The name of the Auto Scaling group.

        resource_id(str, Optional):
            policy name of an autoScalingGroup's policy.

        policy_type(str, Optional):
            One of the following policy types:
                * TargetTrackingScaling
                * StepScaling
                * SimpleScaling (default)
                * PredictiveScaling.

        adjustment_type(str, Optional):
            Specifies how the scaling adjustment is interpreted (for example, an absolute number or a
            percentage). The valid values are ChangeInCapacity, ExactCapacity, and PercentChangeInCapacity.

            Required if the policy type is StepScaling or SimpleScaling. For more information, see Scaling
            adjustment types in the Amazon EC2 Auto Scaling User Guide.
            Defaults to None.

        min_adjustment_step(int, Optional):
            Available for backward compatibility. Use MinAdjustmentMagnitude instead. Defaults to None.

        min_adjustment_magnitude(int, Optional):
            The minimum value to scale by when the adjustment type is PercentChangeInCapacity. For example,
            suppose that you create a step scaling policy to scale out an Auto Scaling group by 25 percent
            and you specify a MinAdjustmentMagnitude of 2. If the group has 4 instances and the scaling
            policy is performed, 25 percent of 4 is 1. However, because you specified a MinAdjustmentMagnitude of 2,
            Amazon EC2 Auto Scaling scales out the group by 2 instances.

            Valid only if the policy type is StepScaling or SimpleScaling. For more information, see Scaling
            adjustment types in the Amazon EC2 Auto Scaling User Guide.

            .. Note::
                Some Auto Scaling groups use instance weights. In this case, set the MinAdjustmentMagnitude to a value
                that is at least as large as your largest instance weight.

            Defaults to None.

        scaling_adjustment(int, Optional):
            The amount by which to scale, based on the specified adjustment type. A positive value adds to the current
            capacity while a negative number removes from the current capacity. For exact capacity, you must specify a
            positive value.

            Required if the policy type is SimpleScaling. (Not used with any other policy type.)
            Defaults to None.

        cooldown(int, Optional):
            The duration of the policy's cooldown period, in seconds. When a cooldown period is specified here, it
            overrides the default cooldown period defined for the Auto Scaling group.

            Valid only if the policy type is SimpleScaling. For more information, see Scaling cooldowns for Amazon EC2
            Auto Scaling in the Amazon EC2 Auto Scaling User Guide.

            Defaults to None.

        metric_aggregation_type(str, Optional):
            The aggregation type for the CloudWatch metrics. The valid values are Minimum, Maximum, and
            Average. If the aggregation type is null, the value is treated as Average.

            Valid only if the policy type is StepScaling. Defaults to None.

        step_adjustments(list[dict[str, Any]], Optional):
            A set of adjustments that enable you to scale based on the size of the alarm breach.

            Required if the policy type is StepScaling. (Not used with any other policy type.) Defaults to None.

            * MetricIntervalLowerBound (float, Optional):
                The lower bound for the difference between the alarm threshold and the CloudWatch metric. If the
                metric value is above the breach threshold, the lower bound is inclusive (the metric must be
                greater than or equal to the threshold plus the lower bound). Otherwise, it is exclusive (the
                metric must be greater than the threshold plus the lower bound). A null value indicates negative
                infinity.

            * MetricIntervalUpperBound (float, Optional):
                The upper bound for the difference between the alarm threshold and the CloudWatch metric. If the
                metric value is above the breach threshold, the upper bound is exclusive (the metric must be
                less than the threshold plus the upper bound). Otherwise, it is inclusive (the metric must be
                less than or equal to the threshold plus the upper bound). A null value indicates positive
                infinity.

                The upper bound must be greater than the lower bound.

            * ScalingAdjustment (int):
                The amount by which to scale, based on the specified adjustment type. A positive value adds to
                the current capacity while a negative number removes from the current capacity.

        estimated_instance_warmup(int, Optional):
            The estimated time, in seconds, until a newly launched instance can contribute to the CloudWatch
            metrics. This warm-up period applies to instances launched due to a specific target tracking or step
            scaling policy. When a warm-up period is specified here, it overrides the default instance warmup.

            Valid only if the policy type is TargetTrackingScaling or StepScaling. Defaults to None.

            .. Note::
                The default is to use the value for the default instance warmup defined for the group. If default
                instance warmup is null, then EstimatedInstanceWarmup falls back to the value of default cooldown.

        target_tracking_configuration(dict[str, Any], Optional):
            A target tracking scaling policy. Provides support for predefined or custom metrics.

            The following predefined metrics are available:
                * ASGAverageCPUUtilization
                * ASGAverageNetworkIn
                * ASGAverageNetworkOut
                * ALBRequestCountPerTarget

            If you specify ALBRequestCountPerTarget for the metric, you must specify the ResourceLabel parameter with
            the PredefinedMetricSpecification.

            For more information, see TargetTrackingConfiguration in the Amazon EC2 Auto Scaling API
            Reference.

            Required if the policy type is TargetTrackingScaling. Defaults to None.

            * PredefinedMetricSpecification (dict[str, Any], Optional):
                A predefined metric. You must specify either a predefined metric or a customized metric.

                * PredefinedMetricType (str):
                    The metric type. The following predefined metrics are available:

                    * ASGAverageCPUUtilization - Average CPU utilization of the Auto Scaling group.

                    * ASGAverageNetworkIn - Average number of bytes received (per instance per minute) for the Auto
                      Scaling group.

                    * ASGAverageNetworkOut - Average number of bytes sent out (per instance per minute) for the Auto
                      Scaling group.

                    * ALBRequestCountPerTarget - Average Application Load Balancer request count (per target per minute)
                      for your Auto Scaling group.

                * ResourceLabel (str, Optional):
                    A label that uniquely identifies a specific Application Load Balancer target group from which to
                    determine the average request count served by your Auto Scaling group. You can't specify a
                    resource label unless the target group is attached to the Auto Scaling group.

                    You create the resource label by appending the final portion of the load balancer ARN and the
                    final portion of the target group ARN into a single value, separated by a forward slash (/). The
                    format of the resource label is: `app/my-alb/778d41231b141a0f/targetgroup/my-alb-target-group/
                    943f017f100becff`.

                    Where:

                    * app/<load-balancer-name>/<load-balancer-id> is the final portion of the load balancer ARN
                    * targetgroup/<target-group-name>/<target-group-id> is the final portion of the target group ARN.

                    To find the ARN for an Application Load Balancer, use the DescribeLoadBalancers API operation. To
                    find the ARN for the target group, use the DescribeTargetGroups API operation.

            * CustomizedMetricSpecification (dict[str, Any], Optional):
                A customized metric. You must specify either a predefined metric or a customized metric.

                * MetricName (str):
                    The name of the metric. To get the exact metric name, namespace, and dimensions, inspect the Metric
                    object that is returned by a call to ListMetrics.

                * Namespace (str):
                    The namespace of the metric.

                * Dimensions (list[dict[str, Any]], Optional):
                    The dimensions of the metric.

                    Conditional: If you published your metric with dimensions, you must specify the same dimensions in
                    your scaling policy.

                    * Name (str):
                        The name of the dimension.
                    * Value (str):
                        The value of the dimension.

                * Statistic (str):
                    The statistic of the metric.

                * Unit (str, Optional):
                    The unit of the metric. For a complete list of the units that CloudWatch supports, see the
                    MetricDatum data type in the Amazon CloudWatch API Reference.

            * TargetValue (float):
                The target value for the metric.

                .. Note::
                    Some metrics are based on a count instead of a percentage, such as the request count for an
                    Application Load Balancer or the number of messages in an SQS queue. If the scaling policy
                    specifies one of these metrics, specify the target utilization as the optimal average request or
                    message count per instance during any one-minute interval.

            * DisableScaleIn (bool, Optional):
                Indicates whether scaling in by the target tracking scaling policy is disabled. If scaling in is
                disabled, the target tracking scaling policy doesn't remove instances from the Auto Scaling
                group. Otherwise, the target tracking scaling policy can remove instances from the Auto Scaling
                group. The default is false.

        enabled(bool, Optional):
            Indicates whether the scaling policy is enabled or disabled. The default is enabled. For more
            information, see Disabling a scaling policy for an Auto Scaling group in the Amazon EC2 Auto
            Scaling User Guide. Defaults to None.

        predictive_scaling_configuration(dict[str, Any], Optional):
            A predictive scaling policy. Provides support for predefined and custom metrics.

            Predefined metrics include CPU utilization, network in/out, and the Application Load Balancer request count.

            For more information, see PredictiveScalingConfiguration in the Amazon EC2 Auto Scaling API Reference.

            Required if the policy type is PredictiveScaling. Defaults to None.

            * MetricSpecifications (list[dict[str, Any]]):
                This structure includes the metrics and target utilization to use for predictive scaling.

                This is an array, but we currently only support a single metric specification. That is, you can
                specify a target value and a single metric pair, or a target value and one scaling metric and
                one load metric.

                * TargetValue (float):
                    Specifies the target utilization.

                    .. Note::
                        Some metrics are based on a count instead of a percentage, such as the request count for an
                        Application Load Balancer or the number of messages in an SQS queue. If the scaling policy
                        specifies one of these metrics, specify the target utilization as the optimal average request
                        or message count per instance during any one-minute interval.

                * PredefinedMetricPairSpecification (dict[str, Any], Optional):
                    The predefined metric pair specification from which Amazon EC2 Auto Scaling determines the
                    appropriate scaling metric and load metric to use.

                    * PredefinedMetricType (str):
                        Indicates which metrics to use. There are two different types of metrics for each metric type:
                        one is a load metric and one is a scaling metric. For example, if the metric type is
                        ASGCPUUtilization, the Auto Scaling group's total CPU metric is used as the load metric, and the
                        average CPU metric is used for the scaling metric.

                    * ResourceLabel (str, Optional):
                        A label that uniquely identifies a specific Application Load Balancer target group from which to
                        determine the total and average request count served by your Auto Scaling group. You can't
                        specify a resource label unless the target group is attached to the Auto Scaling group.

                        You create the resource label by appending the final portion of the load balancer ARN and the
                        final portion of the target group ARN into a single value, separated by a forward slash (/). The
                        format of the resource label is: `app/my-alb/778d41231b141a0f/targetgroup/my-alb-target-group/
                        943f017f100becff`.

                        Where:

                        * app/<load-balancer-name>/<load-balancer-id> is the final portion of the load balancer ARN
                        * targetgroup/<target-group-name>/<target-group-id> is the final portion of the target group ARN

                        To find the ARN for an Application Load Balancer, use the DescribeLoadBalancers API operation.
                        To find the ARN for the target group, use the DescribeTargetGroups API operation.

                * PredefinedScalingMetricSpecification (dict[str, Any], Optional):
                    The predefined scaling metric specification.

                    * PredefinedMetricType (str):
                        The metric type.

                    * ResourceLabel (str, Optional):
                        A label that uniquely identifies a specific Application Load Balancer target group from which to
                        determine the average request count served by your Auto Scaling group. You can't specify a
                        resource label unless the target group is attached to the Auto Scaling group.

                        You create the resource label by appending the final portion of the load balancer ARN and the
                        final portion of the target group ARN into a single value, separated by a forward slash (/).
                        The format of the resource label is: `app/my-alb/778d41231b141a0f/targetgroup/
                        my-alb-target-group/943f017f100becff`.

                        Where:

                        * app/<load-balancer-name>/<load-balancer-id> is the final portion of the load balancer ARN
                        * targetgroup/<target-group-name>/<target-group-id> is the final portion of the target group ARN

                        To find the ARN for an Application Load Balancer, use the DescribeLoadBalancers API operation.
                        To find the ARN for the target group, use the DescribeTargetGroups API operation.

                * PredefinedLoadMetricSpecification (dict[str, Any], Optional):
                    The predefined load metric specification.

                    * PredefinedMetricType (str):
                        The metric type.

                    * ResourceLabel (str, Optional):
                        A label that uniquely identifies a specific Application Load Balancer target group from which to
                        determine the request count served by your Auto Scaling group. You can't specify a resource
                        label unless the target group is attached to the Auto Scaling group.

                        You create the resource label by appending the final portion of the load balancer ARN and the
                        final portion of the target group ARN into a single value, separated by a forward slash (/).
                        The format of the resource label is: `app/my-alb/778d41231b141a0f/targetgroup/my-alb-target-
                        group/943f017f100becff`.

                        Where:

                        * app/<load-balancer-name>/<load-balancer-id> is the final portion of the load balancer ARN
                        * targetgroup/<target-group-name>/<target-group-id> is the final portion of the target group ARN

                        To find the ARN for an Application Load Balancer, use the DescribeLoadBalancers API operation.
                        To find the ARN for the target group, use the DescribeTargetGroups API operation.

                * CustomizedScalingMetricSpecification (dict[str, Any], Optional):
                    The customized scaling metric specification.

                    * MetricDataQueries (list[dict[str, Any]]):
                        One or more metric data queries to provide the data points for a scaling metric. Use multiple
                        metric data queries only if you are performing a math expression on returned data.

                        * Id (str):
                            A short name that identifies the object's results in the response. This name must be unique
                            among all MetricDataQuery objects specified for a single scaling policy. If you are
                            performing math expressions on this set of data, this name represents that data and can
                            serve as a variable in the mathematical expression. The valid characters are letters,
                            numbers, and underscores. The first character must be a lowercase letter.

                        * Expression (str, Optional):
                            The math expression to perform on the returned data, if this object is performing a math
                            expression. This expression can use the Id of the other metrics to refer to those metrics,
                            and can also use the Id of other expressions to use the result of those expressions.

                            Conditional: Within each MetricDataQuery object, you must specify either Expression or
                            MetricStat, but not both.

                        * MetricStat (dict[str, Any], Optional):
                            Information about the metric data to return.

                            Conditional: Within each MetricDataQuery object, you must specify either Expression or
                            MetricStat, but not both.

                            * Metric (dict[str, Any]):
                                The CloudWatch metric to return, including the metric name, namespace, and dimensions.
                                To get the exact metric name, namespace, and dimensions, inspect the Metric object that
                                is returned by a call to ListMetrics.

                                * Namespace (str):
                                    The namespace of the metric. For more information, see the table in Amazon Web
                                    Services services that publish CloudWatch metrics  in the Amazon CloudWatch User
                                    Guide.

                                * MetricName (str):
                                    The name of the metric.

                                * Dimensions (list[dict[str, Any]], Optional):
                                    The dimensions for the metric. For the list of available dimensions, see the Amazon
                                    Web Services documentation available from the table in Amazon Web Services services
                                    that publish CloudWatch metrics  in the Amazon CloudWatch User Guide.

                                    Conditional: If you published your metric with dimensions, you must specify the
                                    same dimensions in your scaling policy.

                                    * Name (str):
                                        The name of the dimension.
                                    * Value (str):
                                        The value of the dimension.

                            * Stat (str):
                                The statistic to return. It can include any CloudWatch statistic or extended statistic.
                                For a list of valid values, see the table in Statistics in the Amazon CloudWatch User
                                Guide.

                                The most commonly used metrics for predictive scaling are Average and Sum.

                            * Unit (str, Optional):
                                The unit to use for the returned data points. For a complete list of the units that
                                CloudWatch supports, see the MetricDatum data type in the Amazon CloudWatch API
                                Reference.

                        * Label (str, Optional):
                            A human-readable label for this metric or expression. This is especially useful if this is a
                            math expression, so that you know what the value represents.

                        * ReturnData (bool, Optional):
                            Indicates whether to return the timestamps and raw data values of this metric.

                            If you use any math expressions, specify true for this value for only the final math
                            expression that the metric specification is based on. You must specify false for ReturnData
                            for all the other metrics and expressions used in the metric specification.

                            If you are only retrieving metrics and not performing any math expressions, do not specify
                            anything for ReturnData. This sets it to its default (true).

                * CustomizedLoadMetricSpecification (dict[str, Any], Optional):
                    The customized load metric specification.

                    * MetricDataQueries (list['MetricDataQuery']):
                        One or more metric data queries to provide the data points for a load metric. Use multiple
                        metric data queries only if you are performing a math expression on returned data.

                * CustomizedCapacityMetricSpecification (dict[str, Any], Optional):
                    The customized capacity metric specification.

                    * MetricDataQueries (list['MetricDataQuery']):
                        One or more metric data queries to provide the data points for a capacity metric. Use multiple
                        metric data queries only if you are performing a math expression on returned data.

            * Mode (str, Optional):
                The predictive scaling mode. Defaults to ForecastOnly if not specified.

            * SchedulingBufferTime (int, Optional):
                The amount of time, in seconds, by which the instance launch time can be advanced. For example,
                the forecast says to add capacity at 10:00 AM, and you choose to pre-launch instances by 5
                minutes. In that case, the instances will be launched at 9:55 AM. The intention is to give
                resources time to be provisioned. It can take a few minutes to launch an EC2 instance. The
                actual amount of time required depends on several factors, such as the size of the instance and
                whether there are startup scripts to complete.

                The value must be less than the forecast interval duration of 3600 seconds (60 minutes). Defaults to
                300 seconds if not specified.

            * MaxCapacityBreachBehavior (str, Optional):
                Defines the behavior that should be applied if the forecast capacity approaches or exceeds the
                maximum capacity of the Auto Scaling group. Defaults to HonorMaxCapacity if not specified.

                The following are possible values:

                * HonorMaxCapacity - Amazon EC2 Auto Scaling cannot scale out capacity higher than the maximum
                  capacity. The maximum capacity is enforced as a hard limit.

                * IncreaseMaxCapacity - Amazon EC2 Auto Scaling can scale out capacity higher than the maximum
                  capacity when the forecast capacity is close to or exceeds the maximum capacity. The upper limit is
                  determined by the forecasted capacity and the value for MaxCapacityBuffer.

            * MaxCapacityBuffer (int, Optional):
                The size of the capacity buffer to use when the forecast capacity is close to or exceeds the
                maximum capacity. The value is specified as a percentage relative to the forecast capacity. For
                example, if the buffer is 10, this means a 10 percent buffer, such that if the forecast capacity
                is 50, and the maximum capacity is 40, then the effective maximum capacity is 55.

                If set to 0, Amazon EC2 Auto Scaling may scale capacity higher than the maximum capacity to equal but
                not exceed forecast capacity.

                Required if the MaxCapacityBreachBehavior property is set to IncreaseMaxCapacity, and cannot be used
                otherwise.

    Request Syntax:
        .. code-block:: sls

            [scaling_policy_name]:
              aws.autoscaling.scaling_policy.present:
                - name: "string"
                - policy_name: "string"
                - auto_scaling_group_name: "string"
                - policy_type: "string"
                - adjustment_type: "string"
                - min_adjustment_step: "int"
                - min_adjustment_magnitude: "int"
                - scaling_adjustment: "int"
                - cooldown: "int"
                - metric_aggregation_type: "string"
                - step_adjustments:
                    - ScalingAdjustment: "int"
                      MetricIntervalLowerBound: "float"
                      MetricIntervalUpperBound: "float"
                - estimated_instance_warmup: "int"
                - target_tracking_configuration:
                    TargetValue: "float"
                    PredefinedMetricSpecification:
                      PredefinedMetricType: "string"
                      ResourceLabel: "string"
                    CustomizedMetricSpecification:
                      MetricName: "string"
                      Namespace: "string"
                      Statistic:
                        Dimensions:
                          - MetricDimension:
                              - Name: "string"
                                Value: "string"
                      Unit: "string"
                    DisableScaleIn: "boolean"
                - enabled: "boolean"
                - predictive_scaling_configuration:
                    MetricSpecifications:
                      - TargetValue: "float"
                        PredefinedMetricPairSpecification:
                          PredefinedMetricType: "string"
                          ResourceLabel: "string"
                        PredefinedScalingMetricSpecification:
                          PredefinedMetricType: "string"
                          ResourceLabel: "string"
                        PredefinedLoadMetricSpecification:
                          PredefinedMetricType: "string"
                          ResourceLabel: "string"
                        CustomizedScalingMetricSpecification:
                          MetricDataQueries:
                            - Id: "string"
                              Expression: "string"
                              MetricStat:
                                Metric:
                                  Namespace: "string"
                                  MetricName: "string"
                                  Dimensions:
                                    - Name: "string"
                                      Value: "string"
                                Stat: "string"
                                Unit: "string"
                              Label: "string"
                              ReturnData: "string"
                        CustomizedLoadMetricSpecification:
                          MetricDataQueries:
                            - Id: "string"
                              Expression: "string"
                              MetricStat:
                                Metric:
                                  Namespace: "string"
                                  MetricName: "string"
                                  Dimensions:
                                    - Name: "string"
                                      Value: "string"
                                Stat: "string"
                                Unit: "string"
                              Label: "string"
                              ReturnData: "string"
                        CustomizedCapacityMetricSpecification:
                          MetricDataQueries:
                            - Id: "string"
                              Expression: "string"
                              MetricStat:
                                Metric:
                                  Namespace: "string"
                                  MetricName: "string"
                                  Dimensions:
                                    - Name: "string"
                                      Value: "string"
                                Stat: "string"
                                Unit: "string"
                              Label: "string"
                              ReturnData: "string"
                    SchedulingBufferTime: "int"
                    MaxCapacityBreachBehavior: "string"
                    MaxCapacityBuffer: "int"


    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            idem_test_autoscaling_policy:
              aws.autoscaling.scaling_policy.present:
                - auto_scaling_group_name: idem-test-autoscaling-group
                - policy_name: idem-test-scaling-policy
                - policy_type: StepScaling
                - adjustment_type: PercentChangeInCapacity
                - min_adjustment_step: 1
                - min_adjustment_magnitude: 1
                - cooldown: 20
                - metric_aggregation_type: Average
                - enabled: true
                - step_adjustments:
                    - MetricIntervalLowerBound: 0.0
                      MetricIntervalUpperBound: 15.0
                      ScalingAdjustment: 1
                    - MetricIntervalLowerBound: 15.0
                      MetricIntervalUpperBound: 25.0
                      ScalingAdjustment: 2
                    - MetricIntervalLowerBound: 25.0
                      ScalingAdjustment: 3

    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    resource_updated = False
    plan_state = None

    if resource_id:
        before = await hub.exec.aws.autoscaling.scaling_policy.get(
            ctx=ctx,
            name=name,
            resource_id=resource_id,
            auto_scaling_group_name=auto_scaling_group_name,
        )
        if not before["result"]:
            result["result"] = False
            result["comment"] = before["comment"]
            return result
        result["old_state"] = copy.deepcopy(before["ret"])
        plan_state = copy.deepcopy(result["old_state"])
        update_ret = await hub.tool.aws.autoscaling.scaling_policy.update(
            ctx=ctx,
            name=name,
            resource_id=resource_id,
            before=result["old_state"],
            auto_scaling_group_name=auto_scaling_group_name,
            policy_type=policy_type,
            adjustment_type=adjustment_type,
            min_adjustment_step=min_adjustment_step,
            min_adjustment_magnitude=min_adjustment_magnitude,
            scaling_adjustment=scaling_adjustment,
            cooldown=cooldown,
            metric_aggregation_type=metric_aggregation_type,
            step_adjustments=step_adjustments,
            estimated_instance_warmup=estimated_instance_warmup,
            target_tracking_configuration=target_tracking_configuration,
            enabled=enabled,
            predictive_scaling_configuration=predictive_scaling_configuration,
        )
        result["comment"] = update_ret["comment"]
        result["result"] = update_ret["result"]
        resource_updated = bool(update_ret["ret"])
        if update_ret["ret"] and ctx.get("test", False):
            for modified_param in update_ret["ret"]:
                plan_state[modified_param] = update_ret["ret"][modified_param]
            result["comment"] += hub.tool.aws.comment_utils.would_update_comment(
                resource_type="aws.autoscaling.scaling_policy", name=name
            )
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "auto_scaling_group_name": auto_scaling_group_name,
                    "policy_type": policy_type,
                    "adjustment_type": adjustment_type,
                    "min_adjustment_step": min_adjustment_step,
                    "min_adjustment_magnitude": min_adjustment_magnitude,
                    "scaling_adjustment": scaling_adjustment,
                    "cooldown": cooldown,
                    "metric_aggregation_type": metric_aggregation_type,
                    "step_adjustments": step_adjustments,
                    "estimated_instance_warmup": estimated_instance_warmup,
                    "target_tracking_configuration": target_tracking_configuration,
                    "enabled": enabled,
                    "predictive_scaling_configuration": predictive_scaling_configuration,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.autoscaling.scaling_policy", name=name
            )
            return result
        ret = await hub.exec.boto3.client.autoscaling.put_scaling_policy(
            ctx,
            AutoScalingGroupName=auto_scaling_group_name,
            PolicyName=name,
            PolicyType=policy_type,
            AdjustmentType=adjustment_type,
            MinAdjustmentStep=min_adjustment_step,
            MinAdjustmentMagnitude=min_adjustment_magnitude,
            ScalingAdjustment=scaling_adjustment,
            Cooldown=cooldown,
            MetricAggregationType=metric_aggregation_type,
            StepAdjustments=step_adjustments,
            EstimatedInstanceWarmup=estimated_instance_warmup,
            TargetTrackingConfiguration=target_tracking_configuration,
            Enabled=enabled,
            PredictiveScalingConfiguration=predictive_scaling_configuration,
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.autoscaling.scaling_policy", name=name
        )
    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not result["old_state"]) or resource_updated:
            after = await hub.exec.aws.autoscaling.scaling_policy.get(
                ctx=ctx,
                name=name,
                resource_id=resource_id,
                auto_scaling_group_name=auto_scaling_group_name,
            )
            if not after["result"]:
                result["result"] = False
                result["comment"] = after["comment"]
                return result
            result["new_state"] = copy.deepcopy(after["ret"])
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(
    hub,
    ctx,
    name: str,
    auto_scaling_group_name: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""Deletes the specified scaling policy.

    Deleting either a step scaling policy or a simple scaling policy deletes the underlying alarm action, but does not
    delete the alarm, even if it no longer has an associated action.

    Args:
        name(str):
            The name or Amazon Resource Name (ARN) of the scaling policy.

        auto_scaling_group_name(str):
            The name of the Auto Scaling group.

        resource_id(str, Optional):
            policy name of an autoScalingGroup's policy.

    Request Syntax:
        .. code-block:: sls

            [rds_scaling_policy_name]:
              aws.autoscaling.scaling_policy.absent:
                - name: 'string'
                - auto_scaling_group_name: 'string'
                - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            rds_scaling_policy_name:
              aws.autoscaling.scaling_policy.absent:
                - name: rds_scaling_policy
                - auto_scaling_group_name: rds_autoscaling_group
                - resource_id: rds_autoscaling_group/rds_autoscaling_policy
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.autoscaling.scaling_policy", name=name
        )
        return result
    before = await hub.exec.aws.autoscaling.scaling_policy.get(
        ctx=ctx,
        name=name,
        resource_id=resource_id,
        auto_scaling_group_name=auto_scaling_group_name,
    )
    if not before["result"]:
        result["result"] = False
        result["comment"] = before["comment"]
        return result
    if not before["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.autoscaling.scaling_policy", name=name
        )
    elif ctx.get("test", False):
        result["old_state"] = before["ret"]
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.autoscaling.scaling_policy", name=name
        )
        return result
    else:
        result["old_state"] = before["ret"]
        ret = await hub.exec.boto3.client.autoscaling.delete_policy(
            ctx,
            PolicyName=resource_id,
            AutoScalingGroupName=auto_scaling_group_name,
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            result["result"] = False
            return result
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.autoscaling.scaling_policy", name=name
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Describes the AutoScaling scaling policies in the account and Region.

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe aws.autoscaling.scaling_policy
    """
    result = {}

    ret = await hub.exec.boto3.client.autoscaling.describe_policies(ctx)
    if not ret["result"]:
        hub.log.debug(
            f"Could not describe aws.autoscaling.scaling_policy. {ret['comment']}"
        )
        return {}
    for scaling_policy in ret["ret"]["ScalingPolicies"]:
        resource_key = f"{scaling_policy.get('AutoScalingGroupName')}/{scaling_policy.get('PolicyName')}"
        resource_translated = hub.tool.aws.autoscaling.conversion_utils.convert_raw_scaling_policy_to_present(
            ctx, raw_resource=scaling_policy
        )
        result[resource_key] = {
            "aws.autoscaling.scaling_policy.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }
    return result



import copy
from typing import Any
from typing import Dict
from typing import List


async def modify_db_group_parameters(
    hub,
    ctx,
    resource_name,
    old_parameters: List[Dict[str, Any]],
    new_parameters: List[Dict[str, Any]],
):
    """
    Update parameters of AWS RDS resources

    Args:
        hub:
        ctx:
        resource_name: aws resource name
        old_parameters: list of old parameters
        new_parameters: list of new parameters

    Returns:
        {"result": True|False, "comment": "A message", "ret": None}

    """
    parameters_to_modify = []
    parameters_to_remove = []
    old_parameters_map = {
        parameter.get("name"): parameter for parameter in old_parameters or []
    }
    parameters_result = copy.deepcopy(old_parameters_map)
    if new_parameters is not None:
        for parameter in new_parameters:
            if parameter.get("name") in old_parameters_map:
                if parameter.get("value") != old_parameters_map.get(
                    parameter.get("name")
                ).get("value"):
                    parameters_to_modify.append(parameter)
                    parameters_to_remove.append(
                        old_parameters_map.get(parameter.get("name"))
                    )
    result = dict(comment=(), result=True, ret=None)
    if not parameters_to_modify:
        return result
    if parameters_to_modify:
        if not ctx.get("test", False):
            update_payload_parameters = []
            for parameter in parameters_to_modify:
                update_payload_parameter = {}
                if parameter["apply_method"] is not None:
                    update_payload_parameter["ApplyMethod"] = parameter["apply_method"]
                if parameter["name"]:
                    update_payload_parameter["ParameterName"] = parameter["name"]
                if parameter["value"]:
                    update_payload_parameter["ParameterValue"] = parameter["value"]
                update_payload_parameters.append(update_payload_parameter)

            update_parameters_ret = (
                await hub.exec.boto3.client.rds.modify_db_parameter_group(
                    ctx,
                    DBParameterGroupName=resource_name,
                    Parameters=update_payload_parameters,
                )
            )
            if not update_parameters_ret["result"]:
                result["comment"] = update_parameters_ret["comment"]
                result["result"] = False
                return result
    for parameter in parameters_to_modify:
        parameters_result[parameter["name"]] = parameter

    result["ret"] = {"parameters": list(parameters_result.values())}
    result["comment"] = result["comment"] + (
        f"Update parameters: Add {[key.get('name') for key in parameters_to_modify]}  Remove {[key.get('name') for key in parameters_to_remove]}",
    )
    return result


async def modify_db_cluster_parameter_group(
    hub,
    ctx,
    resource_name,
    old_parameters: List[Dict[str, Any]],
    new_parameters: List[Dict[str, Any]],
):
    """
    Modifies the parameters of a DB cluster parameter group

    Args:
        hub:
        ctx:
        resource_name: aws resource name
        old_parameters: list of old parameters
        new_parameters: list of new parameters

    Returns:
        {"result": True|False, "comment": "A message", "ret": None}

    """
    parameters_to_modify = []
    old_parameters_map = {
        parameter.get("ParameterName"): parameter for parameter in old_parameters or []
    }

    if new_parameters is not None:
        for parameter in new_parameters:
            if parameter.get("ParameterName") in old_parameters_map:
                if parameter.get("ParameterValue") != old_parameters_map.get(
                    parameter.get("ParameterName")
                ).get("ParameterValue"):
                    parameters_to_modify.append(parameter)

    result = dict(comment=(), result=True, ret=None)
    if not parameters_to_modify:
        return result
    else:
        if not ctx.get("test", False):
            update_parameters_ret = (
                await hub.exec.boto3.client.rds.modify_db_cluster_parameter_group(
                    ctx,
                    DBClusterParameterGroupName=resource_name,
                    Parameters=parameters_to_modify,
                )
            )
            if not update_parameters_ret["result"]:
                result["comment"] = update_parameters_ret["comment"]
                result["result"] = False
                return result

        result["ret"] = {"parameters": parameters_to_modify}
        result["comment"] = result["comment"] + (
            f"Update parameters: Modified {[key.get('ParameterName') for key in parameters_to_modify]}",
        )
    return result



"""State module for managing EC2 placement groups."""
import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    strategy: str = None,
    partition_count: int = None,
    tags: Dict[str, str] = None,
    spread_level: str = None,
) -> Dict[str, Any]:
    """Manage creation and updates for an EC2 placement group.

    Creates a placement group in which to launch instances. The strategy of the placement group determines how the
    instances are organized within the group.  A cluster placement group is a logical grouping of instances within a
    single Availability Zone that benefit from low network latency, high network throughput. A spread placement
    group places instances on distinct hardware. A partition placement group places groups of instances in different
    partitions, where instances in one partition do not share the same hardware with instances in another partition.
    For more information, see Placement groups in the Amazon EC2 User Guide.

    Args:
        name(str):
            An Idem name of the resource and the GroupName of the placement group.

        resource_id(str, Optional):
            The GroupName of the placement group.

        strategy(str, Optional):
            The placement strategy. Defaults to None.
            May be one of the following.
            cluster/spread/partition

        partition_count(int, Optional):
            The number of partitions. Valid only when Strategy is set to partition. Defaults to None.

        tags(dict or list, Optional):
            Dict in the format of ``{tag-key: tag-value}`` or List of tags in the format of
            ``[{"Key": tag-key, "Value": tag-value}]`` to associate with the AMI.

            * Key (str):
                The key name that can be used to look up or retrieve the associated value. For example,
                Department or Cost Center are common choices.

            * Value (str):
                The value associated with this tag. For example, tags with a key name of Department could have
                values such as Human Resources, Accounting, and Support. Tags with a key name of Cost Center
                might have values that consist of the number associated with the different cost centers in your
                company. Typically, many resources have tags with the same key name but with different values.
                Amazon Web Services always interprets the tag Value as a single string. If you need to store an
                array, you can store comma-separated values in the string. However, you must interpret the value
                in your code.

        spread_level(str, Optional):
            Determines how placement groups spread instances.
                * Host: You can use host only with Outpost placement groups.
                * Rack: No usage restrictions. Defaults to None.

    Request Syntax:
        .. code-block:: sls

            idem_placement_group_name:
              aws.ec2.placement_group.present:
                - name: value
                - strategy: cluster

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            idem_placement_group_name:
              aws.ec2.placement_group.present:
                - name: my_placement_group_name
                - strategy: cluster
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False

    if resource_id:
        before = await hub.exec.aws.ec2.placement_group.get(
            ctx, name=name, resource_id=resource_id
        )
        if not before["result"] or not before["ret"]:
            result["result"] = False
            result["comment"] = before["comment"]
            return result

        result["old_state"] = copy.deepcopy(before["ret"])
        plan_state = copy.deepcopy(result["old_state"])
        old_tags = result["old_state"].get("tags")

        if tags is not None and tags != old_tags:
            group_id = before["ret"]["group_id"]
            update_ret = await hub.exec.aws.ec2.tag.update_tags(
                ctx,
                resource_id=group_id,
                old_tags=old_tags,
                new_tags=tags,
            )

            result["comment"] += list(update_ret["comment"])
            result["result"] = update_ret["result"]

            resource_updated = resource_updated or bool(update_ret["ret"])
            if ctx.get("test", False) and update_ret["ret"]:
                plan_state["tags"] = update_ret["ret"]

            if not update_ret["result"]:
                return result

        if ctx.get("test", False) and resource_updated:
            plan_state["tags"] = update_ret["ret"]

            result["new_state"] = plan_state
            result["comment"] += hub.tool.aws.comment_utils.would_update_comment(
                "aws.ec2.placement_group", name
            )
            return result
        if not resource_updated:
            result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
                "aws.ec2.placement_group", name
            )
    else:
        if ctx.get("test", False):
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.ec2.placement_group", name=name
            )
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "resource_id": None,
                    "group_name": name,
                    "strategy": strategy,
                    "partition_count": partition_count,
                    "spread_level": spread_level,
                    "tags": tags,
                },
            )
            return result

        boto_tags = hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
        boto_tag_specification = [
            {"ResourceType": "placement-group", "Tags": boto_tags}
        ]
        ret = await hub.exec.boto3.client.ec2.create_placement_group(
            ctx,
            **{
                "GroupName": name,
                "Strategy": strategy,
                "PartitionCount": partition_count,
                "TagSpecifications": boto_tag_specification,
                "SpreadLevel": spread_level,
            },
        )

        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.ec2.placement_group", name=name
        )

    if ctx.get("test", False):
        result["new_state"] = plan_state
    elif (not before) or resource_updated:
        after = await hub.exec.aws.ec2.placement_group.get(
            ctx, name=name, resource_id=resource_id
        )
        if not after["result"]:
            result["comment"] = after["comment"]
            result["result"] = after["result"]
            return result
        result["new_state"] = copy.deepcopy(after["ret"])
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])

    return result


async def absent(hub, ctx, name: str, resource_id: str) -> Dict[str, Any]:
    """Delete an EC2 placement group if it exists.

    Deletes the specified placement group. You must terminate all instances in the placement group before you can
    delete the placement group. For more information, see Placement groups in the Amazon EC2 User Guide.

    Args:
        name(str): An Idem name of the resource and the GroupName of the placement group
        resource_id(str):
            The GroupName of the placement group.
            Idem automatically considers this resource to be absent if this field is not specified.

    Request Syntax:
        .. code-block:: sls

            idem_placement_group_name:
              aws.ec2.placement_group.absent:
                - name: value
                - resource_id: value

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            idem_placement_group_name:
              aws.ec2.placement_group.absent:
                - name: my_placement_group_name
                - resource_id: my_placement_group_name
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ec2.placement_group", name=name
        )
        return result

    before = await hub.exec.aws.ec2.placement_group.get(
        ctx, name=name, resource_id=resource_id
    )

    if not before["result"]:
        result["comment"] = before["comment"]
        result["result"] = False
        return result

    if not before["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ec2.placement_group", name=name
        )
    elif ctx.get("test", False):
        result["old_state"] = copy.deepcopy(before["ret"])
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.ec2.placement_group", name=name
        )
        return result
    else:
        result["old_state"] = copy.deepcopy(before["ret"])
        ret = await hub.exec.boto3.client.ec2.delete_placement_group(
            ctx, **{"GroupName": resource_id}
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.ec2.placement_group", name=name
        )

    return result


async def describe(
    hub,
    ctx,
) -> Dict[str, Dict[str, Any]]:
    """Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Gets information about the AWS EC2 placement group

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:
        .. code-block:: bash

            $ idem describe aws.ec2.placement_group
    """
    result = {}
    ret = await hub.exec.aws.ec2.placement_group.list(ctx, name=None, filters=None)

    if not ret["result"]:
        hub.log.debug(f"Could not describe aws.ec2.placement_group {ret['comment']}")
        return result

    for placement_group in ret["ret"]:
        resource_id = placement_group.get("group_name", None)
        if resource_id:
            result[resource_id] = {
                "aws.ec2.placement_group.present": [
                    {parameter_key: parameter_value}
                    for parameter_key, parameter_value in placement_group.items()
                ]
            }

    return result



from typing import Any
from typing import Dict


async def get(
    hub,
    ctx,
    *,
    name: str = None,
    user_name: str = None,
    policy_arn: str = None,
    resource_id: str = None,
) -> Dict[str, Any]:
    """
    Check if a managed policy is attached to a user

    Args:
        name(str, optional):
            An Idem name of the state for logging.

        user_name(str, optional):
            The name (friendly name, not ARN) of the IAM user to attach the policy to.
            This parameter allows (through its regex pattern ) a string of characters consisting of upper and lowercase
            alphanumeric characters with no spaces. You can also include any of the following characters: _+=,.@-

        policy_arn(str, optional):
            The Amazon Resource Name (ARN) of the IAM policy you want to attach.

        resource_id(str, optional):
            An identifier refers to an existing resource. The format is <user_name>/<policy_arn> Either resource_id
            or both user_name and policy_arn should be specified for absent.
    """
    result = dict(comment=[], result=True, ret=None)
    if resource_id:
        user_name, policy_arn = resource_id.split("/", 1)
    elif user_name and policy_arn:
        resource_id = f"{user_name}/{policy_arn}"
    else:
        result["result"] = False
        result["comment"] = [
            f"iam.user_policy_attachment {name} either resource_id or both user_name and policy_arn"
            f" should be specified."
        ]
        return result
    user_policies_list = await hub.exec.boto3.client.iam.list_attached_user_policies(
        ctx, UserName=user_name
    )
    if user_policies_list["result"]:
        attached_user_policies_list = user_policies_list["ret"].get("AttachedPolicies")
        if attached_user_policies_list:
            policy_arn_list = [
                policy.get("PolicyArn") for policy in attached_user_policies_list
            ]
            if policy_arn in policy_arn_list:
                result["ret"] = {
                    "name": name,
                    "user_name": user_name,
                    "policy_arn": policy_arn,
                    "resource_id": resource_id,
                }
    else:
        result["comment"] = list(user_policies_list["comment"])
        result["result"] = False
    return result



from typing import Any
from typing import Dict


async def get_caller_identity(hub, ctx) -> Dict[str, Any]:
    """
    Return the caller identity details of IAM user whose credential are used invoke the operation.
    """
    ret = dict(result=True, ret={}, comment="")
    caller_identity = await hub.exec.boto3.client.sts.get_caller_identity(ctx)

    ret["result"] = caller_identity.result
    ret["comment"] = caller_identity.comment
    if caller_identity.result:
        ret["ret"]["UserId"] = caller_identity.ret["UserId"]
        ret["ret"]["Account"] = caller_identity.ret["Account"]
        ret["ret"]["Arn"] = caller_identity.ret["Arn"]
    return ret



import copy
from typing import Any
from typing import Dict


async def update_rest_api(
    hub, ctx, old_state: Dict[str, Any], updateable_resource_parameters: Dict[str, Any]
):
    """
    Updates an AWS API Gateway API resource.

    Args:
        hub: required for functions in hub.
        ctx: context.
        old_state(Dict): Previous state of the rest_api resource
        raw_resource(Dict): Existing resource parameters in Amazon Web Services.
        updateable_resource_parameters(Dict): Parameters from SLS file.

    Returns:
        Dict[str, Any]
    """
    result = dict(comment=(), result=True, ret=None)
    patch_operations = []

    new_state = copy.deepcopy(old_state)
    resource_updated = False
    for key, value in updateable_resource_parameters.items():
        if value is not None:
            if key == "endpoint_configuration":
                if value["types"][0] != old_state["endpoint_configuration"]["types"][0]:
                    patch_operations.append(
                        {
                            "op": "replace",
                            "path": f"/endpointConfiguration/types/{old_state['endpoint_configuration']['types'][0]}",
                            "value": value["types"][0],
                        }
                    )
                    new_state["endpoint_configuration"]["types"][0] = value["types"][0]
                    resource_updated = True

                old_vpc_endpoint_ids = old_state["endpoint_configuration"].get(
                    "vpcEndpointIds", []
                )
                new_vpc_endpoint_ids = value.get("vpcEndpointIds", [])
                if not (
                    hub.tool.aws.state_comparison_utils.are_lists_identical(
                        old_vpc_endpoint_ids,
                        new_vpc_endpoint_ids,
                    )
                ):
                    endpoints_to_remove = list(
                        set(old_vpc_endpoint_ids).difference(set(new_vpc_endpoint_ids))
                    )
                    endpoints_to_add = list(
                        set(new_vpc_endpoint_ids).difference(set(old_vpc_endpoint_ids))
                    )
                    if endpoints_to_remove:
                        patch_operations.append(
                            {
                                "op": "remove",
                                "path": f"/endpointConfiguration/vpcEndpointIds",
                                "value": ",".join(endpoints_to_remove),
                            }
                        )
                    if endpoints_to_add:
                        patch_operations.append(
                            {
                                "op": "add",
                                "path": f"/endpointConfiguration/vpcEndpointIds",
                                "value": ",".join(endpoints_to_add),
                            }
                        )

                    new_state["endpoint_configuration"][
                        "vpcEndpointIds"
                    ] = new_vpc_endpoint_ids
                    resource_updated = True
            elif (
                isinstance(value, list)
                and not hub.tool.aws.state_comparison_utils.are_lists_identical(
                    value,
                    old_state.get(key, []),
                )
                or value != old_state[key]
            ):
                if old_state.get(key):
                    patch_operations.append(
                        {
                            "op": "replace",
                            "path": f"/{key}",
                            "value": str(value)
                            if not isinstance(value, list)
                            else ",".join(value),
                        }
                    )
                else:
                    patch_operations.append(
                        {
                            "op": "add",
                            "path": f"/{key}",
                            "value": str(value)
                            if not isinstance(value, list)
                            else ",".join(value),
                        }
                    )
                new_state[key] = value
                resource_updated = True
    if ctx.get("test", False):
        if resource_updated:
            result["comment"] = hub.tool.aws.comment_utils.would_update_comment(
                resource_type="aws.apigateway.rest_api", name=old_state["name"]
            )
            result["ret"] = new_state
        else:
            result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
                resource_type="aws.apigateway.rest_api", name=old_state["name"]
            )
        return result

    if resource_updated:
        update_ret = await hub.exec.boto3.client.apigateway.update_rest_api(
            ctx, restApiId=old_state["resource_id"], patchOperations=patch_operations
        )
        if not update_ret["result"]:
            result["result"] = False
            result["comment"] = update_ret["comment"]
            return result
        else:
            result["comment"] = hub.tool.aws.comment_utils.update_comment(
                resource_type="aws.apigateway.rest_api", name=old_state["name"]
            )
            result["ret"] = new_state

    else:
        result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
            resource_type="aws.apigateway.rest_api", name=old_state["name"]
        )

    return result



import copy
from typing import Any
from typing import Dict


async def update_tags(
    hub,
    ctx,
    resource_id: str,
    old_tags: Dict[str, Any],
    new_tags: Dict[str, Any],
):
    """

    Args:
        resource_id: aws organizations resource id
        old_tags: dict in the format of {tag-key: tag-value}
        new_tags: dict in the format of {tag-key: tag-value}. If this value is None, the function will do no operation on tags.


    Returns:
        {"result": True|False, "comment": "A message Tuple", "ret": dict of updated tags}

    """

    tags_to_add = {}
    tags_to_remove = {}
    if new_tags is not None:
        tags_to_remove, tags_to_add = hub.tool.aws.tag_utils.diff_tags_dict(
            old_tags=old_tags, new_tags=new_tags
        )
    result = dict(comment=(), result=True, ret=None)
    if (not tags_to_remove) and (not tags_to_add):
        result["ret"] = copy.deepcopy(old_tags if old_tags else {})
        return result
    if tags_to_remove:
        if not ctx.get("test", False):
            delete_tag_resp = await hub.exec.boto3.client.organizations.untag_resource(
                ctx, ResourceId=resource_id, TagKeys=list(tags_to_remove.keys())
            )
            if not delete_tag_resp["result"]:
                hub.log.debug(
                    f"Could not delete tags {tags_to_remove} on resource {resource_id} with error {delete_tag_resp['comment']}"
                )
                result["comment"] = result["comment"] + delete_tag_resp["comment"]
                result["result"] = delete_tag_resp["result"]
                return result
    if tags_to_add:
        if not ctx.get("test", False):
            create_tag_resp = await hub.exec.boto3.client.organizations.tag_resource(
                ctx,
                ResourceId=resource_id,
                Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags_to_add),
            )
            if not create_tag_resp["result"]:
                hub.log.debug(
                    f"Could not create tags {tags_to_add} on resource {resource_id} with error {create_tag_resp['comment']}"
                )
                result["comment"] = result["comment"] + create_tag_resp["comment"]
                result["result"] = create_tag_resp["result"]
                return result
    result["ret"] = new_tags
    if ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_update_tags_comment(
            tags_to_remove=tags_to_remove, tags_to_add=tags_to_add
        )
    else:
        result["comment"] = hub.tool.aws.comment_utils.update_tags_comment(
            tags_to_remove=tags_to_remove, tags_to_add=tags_to_add
        )
    return result



from collections import OrderedDict
from typing import Any
from typing import Dict

"""
 Utility functions to support aws efs file system operations
"""


def convert_raw_file_system_to_present(
    hub, file_system: Dict[str, Any], backup_enabled: bool
) -> Dict[str, Any]:
    result = {}
    if file_system is None:
        return None

    resource_parameters = OrderedDict(
        {
            "PerformanceMode": "performance_mode",
            "Encrypted": "encrypted",
            "KmsKeyId": "kms_key_id",
            "CreationToken": "creation_token",
            "ThroughputMode": "throughput_mode",
            "ProvisionedThroughputInMibps": "provisioned_throughput_in_mibps",
            "AvailabilityZoneName": "availability_zone_name",
        }
    )

    tag_parameters = OrderedDict({"Key": "key", "Value": "value"})
    # Populate state parameters
    result["name"] = file_system.get("Name")
    result["resource_id"] = file_system.get("FileSystemId")

    for parameter_key, parameter_value in resource_parameters.items():
        if parameter_key in file_system:
            result[parameter_value] = file_system.get(parameter_key)

    # Special handling for backup
    result["backup"] = backup_enabled

    # Special handling for tags
    translated_tags = []
    for tag in file_system.get("Tags"):
        if tag.get("Key") == "aws:elasticfilesystem:default-backup":
            continue
        translated_tags.append(tag)
    result["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(translated_tags)
    return result


def is_throughput_mode_updated(
    hub,
    current_state: Dict[str, Any],
    desired_state: Dict[str, Any],
) -> bool:
    return current_state.get("throughput_mode") != desired_state.get(
        "throughput_mode"
    ) or (
        current_state.get("throughput_mode") == "provisioned"
        and current_state.get("provisioned_throughput_in_mibps")
        != desired_state.get("provisioned_throughput_in_mibps")
    )


def is_backup_enabled(hub, backup_result):
    return (
        True
        if backup_result["result"]
        and backup_result["ret"]["BackupPolicy"]["Status"] in ["ENABLED", "ENABLING"]
        else False
    )


def is_backup_policy_not_found(hub, backup_result):
    return bool(
        [item for item in backup_result["comment"] if item.startswith("PolicyNotFound")]
    )



"""State module for managing Lambda provisioned concurrency config."""
import copy
from typing import Any

__contracts__ = ["resource"]

from typing import Dict


async def present(
    hub,
    ctx,
    name: str,
    function_name: str,
    qualifier: str,
    provisioned_concurrent_executions: int,
    resource_id: str = None,
) -> Dict[str, Any]:
    """Adds a provisioned concurrency configuration to a function's alias or version.

    Args:
        name(str): An Idem name of the resource.
        function_name(str):
            The name of the Lambda function. The length constraint applies only to the full ARN. If you  specify only the function name, it is limited to 64 characters in length.
                Name formats:
                    * Function name - my-function.
                    * Function ARN - ``arn:aws:lambda:us-west-2:123456789012:function:my-function`` .
                    * Partial ARN - ``123456789012:function:my-function`` .
        qualifier(str): A version number or alias name. Defaults to None.
        provisioned_concurrent_executions(int): The amount of provisioned concurrency to allocate for the version or alias.
        resource_id(str, Optional): The Amazon Resource Name (ARN) of the function. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            function_provisioned_concurrency_config-01234672f3336db8:
              aws.lambda_aws.function_provisioned_concurrency_config.present:
                - name: value
                - function_name: value
                - qualifier: value
                - provisioned_concurrent_executions: int
    """
    result = dict(comment=[], old_state=None, new_state=None, name=name, result=True)

    before = None
    resource_updated = False
    if resource_id:
        before = (
            await hub.exec.aws.lambda_aws.function_provisioned_concurrency_config.get(
                ctx, name=name, resource_id=resource_id
            )
        )
        if not (before["result"] and before["ret"]):
            result["comment"] = before["comment"]
            result["result"] = False
            return result
        result["old_state"] = copy.deepcopy(before["ret"])
        result["new_state"] = copy.deepcopy(result["old_state"])

        old_provisioned_concurrent_executions = result["old_state"].get(
            "provisioned_concurrent_executions"
        )
        if old_provisioned_concurrent_executions != provisioned_concurrent_executions:
            if ctx.get("test", False):
                result["comment"] = hub.tool.aws.comment_utils.would_update_comment(
                    resource_type="aws.lambda.function_provisioned_concurrency_config",
                    name=name,
                )
                result["new_state"][
                    "provisioned_concurrent_executions"
                ] = provisioned_concurrent_executions
                return result
            else:
                update_ret = await hub.exec.boto3.client[
                    "lambda"
                ].put_provisioned_concurrency_config(
                    ctx,
                    FunctionName=function_name,
                    Qualifier=qualifier,
                    ProvisionedConcurrentExecutions=provisioned_concurrent_executions,
                )
                if not update_ret["result"]:
                    result["result"] = False
                    result["comment"] = update_ret["comment"]
                    return result
                result["comment"] = hub.tool.aws.comment_utils.update_comment(
                    resource_type="aws.lambda.function_provisioned_concurrency_config",
                    name=name,
                )
                resource_updated = True
        else:
            result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
                resource_type="aws.lambda.function_provisioned_concurrency_config",
                name=name,
            )

    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "function_name": function_name,
                    "qualifier": qualifier,
                    "provisioned_concurrent_executions": provisioned_concurrent_executions,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.lambda.function_provisioned_concurrency_config",
                name=name,
            )
            return result

        create_ret = await hub.exec.boto3.client[
            "lambda"
        ].put_provisioned_concurrency_config(
            ctx,
            FunctionName=function_name,
            Qualifier=qualifier,
            ProvisionedConcurrentExecutions=provisioned_concurrent_executions,
        )
        result["result"] = create_ret["result"]
        if not result["result"]:
            result["comment"] = create_ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.lambda.function_provisioned_concurrency_config",
            name=name,
        )
    if (not before) or resource_updated:
        after = (
            await hub.exec.aws.lambda_aws.function_provisioned_concurrency_config.get(
                ctx, name=name, function_name=function_name, qualifier=qualifier
            )
        )
        if not (after["result"] and after["ret"]):
            result["result"] = False
            result["comment"] = result["comment"] + after["comment"]
            return result
        result["new_state"] = copy.deepcopy(after["ret"])

    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    """Deletes the provisioned concurrency configuration for a function.

    Args:
        name(str): An Idem name of the resource.
        resource_id(str, Optional): The Amazon Resource Name (ARN) of the function. Idem automatically considers this
            resource being absent if this field is not specified.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            function_provisioned_concurrency_config-01234672f3336db8:
              aws.lambda_aws.function_provisioned_concurrency_config.absent:
                - name: value
                - resource_id: value
    """
    already_absent_msg = hub.tool.aws.comment_utils.already_absent_comment(
        resource_type="aws.lambda.function_provisioned_concurrency_config", name=name
    )
    result = dict(
        comment=already_absent_msg,
        old_state=None,
        new_state=None,
        name=name,
        result=True,
    )

    if not resource_id:
        return result

    before = await hub.exec.aws.lambda_aws.function_provisioned_concurrency_config.get(
        ctx, name=name, resource_id=resource_id
    )

    if not before["result"]:
        result["comment"] = before["comment"]
        result["result"] = False
        return result

    if before["ret"]:
        result["old_state"] = before["ret"]
        if ctx.get("test", False):
            result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
                resource_type="aws.lambda.function_provisioned_concurrency_config",
                name=name,
            )
        else:
            function_name = hub.tool.aws.arn_utils.get_resource_name(resource_id)
            qualifier = hub.tool.aws.arn_utils.get_qualifier(resource_id)
            ret = await hub.exec.boto3.client[
                "lambda"
            ].delete_provisioned_concurrency_config(
                ctx, **{"FunctionName": function_name, "Qualifier": qualifier}
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = hub.tool.aws.comment_utils.delete_comment(
                resource_type="aws.lambda.function_provisioned_concurrency_config",
                name=name,
            )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Retrieves a list of provisioned concurrency configurations for a function.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe aws.lambda_aws.function_provisioned_concurrency_config
    """
    result = {}

    get_functions_ret = await hub.exec.boto3.client["lambda"].list_functions(ctx)
    if not get_functions_ret["result"]:
        hub.log.debug(f"Could not list functions {get_functions_ret['comment']}")
        return result

    for function in get_functions_ret["ret"]["Functions"]:
        function_name = function.get("FunctionName")
        list_concurrency_configs_ret = (
            await hub.exec.aws.lambda_aws.function_provisioned_concurrency_config.list(
                ctx, function_name=function_name
            )
        )

        if not list_concurrency_configs_ret["result"]:
            hub.log.debug(
                f"Could not list provisioned concurrency configs for functionName '{function_name}': "
                f"{list_concurrency_configs_ret['comment']}. Describe will skip this function and continue."
            )
            continue

        for function_concurrency_config in list_concurrency_configs_ret["ret"]:
            # We set function ARN as name of the resource when performing describe
            resource_arn = function_concurrency_config.get("name")
            result[resource_arn] = {
                "aws.lambda.function_provisioned_concurrency_config.present": [
                    {parameter_key: parameter_value}
                    for parameter_key, parameter_value in function_concurrency_config.items()
                ]
            }
    return result
